"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
/**
 * A scope class that contains symbols in the current scope
 */
class Environment {
    constructor() {
        this.varTable = new Map();
        this.funcTable = new Map();
    }
    /**
     * Get a value from the environment, first functions then variables
     * @param key key lookup
     */
    get(key) {
        return this.funcTable.get(key) || this.varTable.get(key);
    }
    /**
     * Get a value from the environment, functions or variables
     * @param key key lookup
     */
    getKind(key, kind) {
        // check functions then variables
        switch (kind) {
            case types_1.KsSymbolKind.variable:
            case types_1.KsSymbolKind.parameter:
                return this.varTable.get(key);
            case types_1.KsSymbolKind.function:
            case types_1.KsSymbolKind.lock:
                return this.funcTable.get(key);
            default:
                throw new Error('Unexpected symbol kind');
        }
    }
    /**
     * See if an environment has a value, first functions then variables
     * @param key key lookup
     */
    has(key) {
        // check functions then variables
        return this.funcTable.has(key) || this.varTable.has(key);
    }
    /**
     * See if an environment has a value, functions or variables
     * @param key key lookup
     */
    hasKind(key, kind) {
        // check functions then variables
        switch (kind) {
            case types_1.KsSymbolKind.variable:
            case types_1.KsSymbolKind.parameter:
                return this.varTable.has(key);
            case types_1.KsSymbolKind.function:
            case types_1.KsSymbolKind.lock:
                return this.funcTable.has(key);
            default:
                throw new Error('Unexpected symbol kind');
        }
    }
    /**
     * Set a value to the environment
     * @param key key lookup
     * @param value basic tracker
     */
    set(key, kind, value) {
        switch (kind) {
            case types_1.KsSymbolKind.variable:
            case types_1.KsSymbolKind.parameter:
                return this.varTable.set(key, value);
            case types_1.KsSymbolKind.function:
            case types_1.KsSymbolKind.lock:
                return this.funcTable.set(key, value);
            default:
                throw new Error('Environment cannot hold suffixes');
        }
    }
    /**
     * Get all trackers in this environment
     */
    trackers() {
        const symbols = [];
        // get variable and parameter symbols
        for (const trackers of this.varTable.values()) {
            symbols.push(trackers);
        }
        // get function and lock symbols
        for (const trackers of this.funcTable.values()) {
            symbols.push(trackers);
        }
        return symbols;
    }
    /**
     * All the symbols in this environment
     */
    symbols() {
        const symbols = [];
        // get variable and parameter symbols
        for (const trackers of this.varTable.values()) {
            symbols.push(trackers.declared.symbol);
        }
        // get function and lock symbols
        for (const trackers of this.funcTable.values()) {
            symbols.push(trackers.declared.symbol);
        }
        return symbols;
    }
}
exports.Environment = Environment;
//# sourceMappingURL=environment.js.map