"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const positionUtils_1 = require("../utilities/positionUtils");
const logger_1 = require("../utilities/logger");
const typeGuards_1 = require("../utilities/typeGuards");
const types_1 = require("../parser/types");
const constants_1 = require("../utilities/constants");
/**
 * The Symbol table is used to update and retrieve symbol type information
 */
class SymbolTable {
    /**
     * construct a symbol table. typically this is done by the symbol table builder
     * @param rootScope the root scope of the symbol table
     * @param dependencyTables tables that this table is dependent on
     * @param dependentTable tables that are dependent on this table
     * @param uri the file uri for this symbol table
     * @param logger a logger for the symbol table
     */
    constructor(rootScope, dependencyTables, dependentTable, uri, logger = logger_1.mockLogger) {
        this.rootScope = rootScope;
        this.dependencyTables = dependencyTables;
        this.dependentTables = dependentTable;
        this.uri = uri;
        this.logger = logger;
        for (const symbolTable of dependencyTables) {
            symbolTable.dependentTables.add(this);
        }
    }
    /**
     * return itself for the graph interface
     */
    get value() {
        return this;
    }
    /**
     * get all adjacent nodes to this symbol table
     */
    get adjacentNodes() {
        return Array.from(this.dependencyTables);
    }
    /**
     * should be called when the associated file is closed. It will remove it's reference
     * if another symbol table doesn't reference it
     */
    closeSelf() {
        // remove children if no parents
        if (this.dependentTables.size === 0) {
            // remove references from child scopes
            for (const child of this.dependencyTables) {
                child.dependentTables.delete(this);
            }
            // clear own references
            this.dependencyTables.clear();
        }
    }
    /**
     * This should be called when the associated symbol table is to be deleted
     */
    removeSelf() {
        // remove references from parent scopes
        for (const parent of this.dependentTables) {
            parent.dependencyTables.delete(this);
        }
        // remove references from child scopes
        for (const child of this.dependencyTables) {
            child.dependentTables.delete(this);
        }
        // clear own references
        this.dependencyTables.clear();
        this.dependentTables.clear();
    }
    /**
     * get every symbol in this file
     */
    allSymbols() {
        const symbols = [];
        this.fileSymbolsDepth(symbols, this.rootScope);
        return symbols;
    }
    /**
     * Get global symbols that match the requested symbol name
     * @param name symbol name
     * @param has function to determine if an environment has the appropriate symbol
     * @param state are we currently looking up or down the dependency tree
     */
    globalEnvironment(name, state, searched, has = (env, lookup) => env.has(lookup)) {
        for (const environment of this.importedEnvironments_(searched, state)) {
            if (has(environment, name)) {
                return environment;
            }
        }
        return undefined;
    }
    /**
     * Get all symbols available at the provide position
     * @param pos position to check
     */
    scopedSymbols(pos) {
        return this.scopedTrackers(pos).map(tracker => tracker.declared.symbol);
    }
    /**
     * Get a tracker for a given symbol under a filter condition and position
     * @param pos position to check
     * @param name name of the symbol
     * @param symbolFilter filter function for the symbol
     */
    scopedNamedTracker(pos, name, symbolFilter) {
        // our base filter is to match the symbol lexeme
        const baseFilter = (trackers) => trackers.declared.symbol.name.lookup === name;
        // our compound filter checks for any and the other requested filtering operations
        const compoundFilter = typeGuards_1.empty(symbolFilter)
            ? baseFilter
            : (trackers) => baseFilter(trackers) && symbolFilter(trackers);
        return this.scopedTracker(pos, compoundFilter);
    }
    /**
     * recursively move up the scope to get every file symbol
     * @param symbols symbols found so far
     * @param node current environment node
     */
    fileSymbolsDepth(symbols, node) {
        symbols.push(...node.environment.symbols());
        for (const child of node.children) {
            this.fileSymbolsDepth(symbols, child);
        }
    }
    /**
     * get all symbol trackers in scope at a position
     * @param pos the position to retrieve symbols from
     */
    scopedTrackers(pos) {
        const trackers = [];
        this.scopedTrackersDepth(trackers, pos, this.rootScope);
        trackers.push(...this.importedTrackers());
        return trackers;
    }
    /**
     * get a symbol tracker under the conditions of a provided filter
     * @param pos the position to check for symbols
     * @param trackerFilter the tracker filter
     */
    scopedTracker(pos, trackerFilter = _ => true) {
        const tracker = this.scopedTrackerDepth(pos, this.rootScope, trackerFilter);
        if (!typeGuards_1.empty(tracker)) {
            return tracker;
        }
        // search all imported trackers
        for (const tracker of this.importedTrackers()) {
            if (trackerFilter(tracker)) {
                return tracker;
            }
        }
        return undefined;
    }
    /**
     * recursively move up scopes to find most relevant symbols
     * @param trackers basic trackers found so far
     * @param pos position to search
     * @param nodes scope nodes to check
     */
    scopedTrackersDepth(trackers, pos, node) {
        const { range } = node;
        switch (range.kind) {
            case types_1.ScopeKind.global:
                // if global scope we're in scope
                trackers.push(...node.environment.trackers());
                for (const child of node.children) {
                    this.scopedTrackersDepth(trackers, pos, child);
                }
                break;
            case types_1.ScopeKind.local:
                // if local scope check that we're in range
                if (positionUtils_1.rangeContainsPos(range, pos)) {
                    trackers.push(...node.environment.trackers());
                    for (const child of node.children) {
                        this.scopedTrackersDepth(trackers, pos, child);
                    }
                    break;
                }
        }
    }
    /**
     * recursively move up scope for more relevant symbols
     * @param pos position to search
     * @param nodes scope nodes to check
     * @param trackerFilter filter for acceptable symbols
     */
    scopedTrackerDepth(pos, node, trackerFilter) {
        let childSymbol = undefined;
        const { range } = node;
        switch (range.kind) {
            // if global it is available
            case types_1.ScopeKind.global:
                for (const child of node.children) {
                    childSymbol = this.scopedTrackerDepth(pos, child, trackerFilter);
                    if (!typeGuards_1.empty(childSymbol)) {
                        return childSymbol;
                    }
                }
                for (const tracker of node.environment.trackers()) {
                    if (trackerFilter(tracker)) {
                        return tracker;
                    }
                }
                break;
            // if the scope has a real position check if we're in the bounds
            case types_1.ScopeKind.local:
                if (positionUtils_1.rangeContainsPos(range, pos)) {
                    for (const child of node.children) {
                        childSymbol = this.scopedTrackerDepth(pos, child, trackerFilter);
                        if (!typeGuards_1.empty(childSymbol)) {
                            return childSymbol;
                        }
                    }
                    for (const tracker of node.environment.trackers()) {
                        if (trackerFilter(tracker)) {
                            return tracker;
                        }
                    }
                }
                break;
        }
        return undefined;
    }
    /**
     * Search all imported trackers that are accessible from this
     * environment
     */
    *importedTrackers() {
        for (const environment of this.importedEnvironments()) {
            yield* environment.trackers();
        }
    }
    /**
     * Search all accessible environments from this environment
     * @param checked what environments have already been checked
     * @param state what is the current search state
     */
    *importedEnvironments() {
        const checked = new Set([this]);
        // If we're currently looking up the dependency tree we can continue doing so
        for (const dependent of this.dependentTables) {
            yield* dependent.importedEnvironments_(checked, 0 /* dependents */);
        }
        // check dependencies
        for (const dependency of this.dependencyTables) {
            yield* dependency.importedEnvironments_(checked, 1 /* dependencies */);
        }
    }
    /**
     * Search all accessible environments from this environment
     * @param checked what environments have already been checked
     * @param state what is the current search state
     */
    *importedEnvironments_(checked, state) {
        // return if we already checked this environment
        if (checked.has(this)) {
            return;
        }
        checked.add(this);
        yield this.rootScope.environment;
        // don't check any further if built in
        if (this.uri === constants_1.builtIn) {
            return;
        }
        // If we're currently looking up the dependency tree we can continue doing so
        if (state === 0 /* dependents */) {
            for (const dependent of this.dependentTables) {
                yield* dependent.importedEnvironments_(checked, 0 /* dependents */);
            }
        }
        // check dependencies
        for (const dependency of this.dependencyTables) {
            yield* dependency.importedEnvironments_(checked, 1 /* dependencies */);
        }
    }
}
exports.SymbolTable = SymbolTable;
//# sourceMappingURL=symbolTable.js.map