import { Position, Location, Range, Connection } from 'vscode-languageserver';
import { KLSConfiguration } from './types';
import { KsSymbol, SymbolTracker, KsBaseSymbol } from './analysis/types';
import { Token } from './entities/token';
export declare class KLS {
    /**
     * What is the workspace uri
     */
    workspaceUri?: string;
    /**
     * The logger used by this and all dependencies
     */
    private readonly logger;
    /**
     * The tracer used by this and all dependencies
     */
    private readonly tracer;
    /**
     * Connection to the client
     */
    private readonly connection;
    /**
     * This server's configuration
     */
    private readonly configuration;
    /**
     * The document service to store and manage documents
     */
    private readonly documentService;
    /**
     * A service to take document info and generate all foldable regions
     */
    private readonly foldableService;
    /**
     * A service to provide analysis against scripts
     */
    private readonly analysisService;
    constructor(caseKind: CaseKind | undefined, logger: ILogger | undefined, tracer: ITracer | undefined, connection: Connection, configuration: KLSConfiguration);
    /**
     * Start the language server listening to requests from the client
     */
    listen(): void;
    /**
     * Initialize the server from the client connection
     * @param params initialization parameters
     */
    private onInitialize;
    /**
     * Post initialization register additional hooks and retrieve client
     * configurations
     * @param _ initialized parameters
     */
    private onInitialized;
    /**
     * Update the server configuration when the client signals a change in it's configuration for
     * the kos-language-server
     * @param change The updated settings
     */
    private onDidChangeConfiguration;
    /**
     * Respond to completion requests from the client. This handler currently provides
     * both symbol completion as well as suffix completion.
     * @param completion the parameters describing the completion request
     */
    private onCompletion;
    /**
     * This handler provider completion item resolution capability. This provides
     * additional information for the currently completion item selection
     * @param completionItem the item to resolve further
     */
    private onCompletionResolve;
    /**
     * This handler provider rename capabilities. This allows a client to highlight
     * as symbol and provide a new name that will change for all known symbols
     * @param rename information describing what and where a rename should occur
     */
    private onRenameRequest;
    /**
     * This handler provides highlight within a requested document. This allows the client
     * to highlight and symbol and other instances of that symbol to also highlight.
     * @param positionParams the position of the highlight request
     */
    private onDocumentHighlight;
    /**
     * This handler provides on hover capability for symbols in a document. This allows additional
     * information to be displayed to the user about symbols throughout the document
     * @param positionParams the position of the hover request
     */
    private onHover;
    /**
     * This handler provides reference capabilities to symbols in a document. This allows a client
     * to identify all positions that a symbol is used in the document or attached documents
     * @param reference parameters describing the reference request
     */
    private onReference;
    /**
     * This handler provides signature help suffixes and function within the document. This
     * provides extra context to the client such as the current parameter
     * @param positionParams the position of the signature request
     */
    private onSignatureHelp;
    /**
     * This handler provides document symbol capabilities. This provides a list of all
     * symbols that are located within a given document
     * @param documentSymbol the document to provide symbols for
     */
    private onDocumentSymbol;
    /**
     * This handler provides go to definition capabilities. When a client requests a symbol
     * go to definition this provides the location if it exists
     * @param positionParams the position of the definition request
     */
    private onDefinition;
    /**
     * This handler provide folding region capabilities. The client will ask for available folding
     * region in which this will respond with the ranges defined by #region and #endregion
     * @param foldingParams the document to preform folding analysis on
     */
    private onFoldingRange;
    /**
     * Respond to updates made to document by the client. This method
     * will parse and update the internal state of affects scripts
     * reporting errors to the client as they are discovered
     * @param document the updated document
     */
    private onChange;
    /**
     * Get document settings from the client. If the client does not support
     * have configurations then return the default configurations.
     */
    private getDocumentSettings;
    /**
     * Update the servers configuration in response to a change in the client configuration
     * @param clientConfig client configuration
     */
    private updateServer;
    /**
     * Set the volume 0 path for the analyzer
     * @param uri path of volume 0
     */
    private setUri;
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind: CaseKind): void;
    /**
     * Get the token at the provided position in the text document
     * @param pos position in the text document
     * @param uri uri of the text document
     */
    getToken(pos: Position, uri: string): Promise<Maybe<Token>>;
    /**
     * Get the declaration location for the token at the provided position
     * @param pos position in the document
     * @param uri uri of the document
     */
    getDeclarationLocation(pos: Position, uri: string): Promise<Maybe<Location>>;
    /**
     * Get all usage locations in all files
     * @param pos position in document
     * @param uri uri of document
     */
    getUsageLocations(pos: Position, uri: string): Promise<Maybe<Location[]>>;
    /**
     * Get all usage ranges in a provide file
     * @param pos position in document
     * @param uri uri of document
     */
    getFileUsageRanges(pos: Position, uri: string): Promise<Maybe<Range[]>>;
    /**
     * Get all symbols in scope at a particular location in the file
     * @param pos position in document
     * @param uri document uri
     */
    getScopedSymbols(pos: Position, uri: string): Promise<KsBaseSymbol[]>;
    /**
     * Get all symbols in a provided file
     * @param uri document uri
     */
    getAllFileSymbols(uri: string): Promise<KsSymbol[]>;
    /**
     * Get a function located at the current location
     * @param pos position in the document
     * @param uri document uri
     */
    getFunctionAtPosition(pos: Position, uri: string): Promise<Maybe<{
        tracker: SymbolTracker;
        index: number;
    }>>;
}
