"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const scanner_1 = require("./scanner/scanner");
const types_1 = require("./analysis/types");
const logger_1 = require("./utilities/logger");
const typeGuards_1 = require("./utilities/typeGuards");
const scriptFind_1 = require("./parser/scriptFind");
const Expr = __importStar(require("./parser/expr"));
const Stmt = __importStar(require("./parser/stmt"));
const SuffixTerm = __importStar(require("./parser/suffixTerm"));
const constants_1 = require("./utilities/constants");
const positionUtils_1 = require("./utilities/positionUtils");
const vscode_uri_1 = require("vscode-uri");
const documentService_1 = require("./services/documentService");
const serverUtils_1 = require("./utilities/serverUtils");
const clean_1 = require("./utilities/clean");
const tokentypes_1 = require("./entities/tokentypes");
const typeUtilities_1 = require("./typeChecker/typeUtilities");
const documentLoader_1 = require("./utilities/documentLoader");
const foldableService_1 = require("./services/foldableService");
const analysisService_1 = require("./services/analysisService");
const pathResolver_1 = require("./utilities/pathResolver");
const fs_1 = require("fs");
class KLS {
    constructor(caseKind = 2 /* camelcase */, logger = logger_1.mockLogger, tracer = logger_1.mockTracer, connection, configuration) {
        this.workspaceUri = undefined;
        this.logger = logger;
        this.tracer = tracer;
        this.configuration = configuration;
        this.connection = connection;
        this.documentService = new documentService_1.DocumentService(connection, new documentLoader_1.DocumentLoader(), logger, tracer);
        this.foldableService = new foldableService_1.FoldableService();
        this.analysisService = new analysisService_1.AnalysisService(caseKind, this.logger, this.tracer, this.documentService);
    }
    /**
     * Start the language server listening to requests from the client
     */
    listen() {
        this.connection.onInitialize(this.onInitialize.bind(this));
        this.connection.onInitialized(this.onInitialized.bind(this));
        this.connection.onDidChangeConfiguration(this.onDidChangeConfiguration.bind(this));
        this.connection.onCompletion(this.onCompletion.bind(this));
        this.connection.onCompletionResolve(this.onCompletionResolve.bind(this));
        this.connection.onRenameRequest(this.onRenameRequest.bind(this));
        this.connection.onDocumentHighlight(this.onDocumentHighlight.bind(this));
        this.connection.onHover(this.onHover.bind(this));
        this.connection.onReferences(this.onReference.bind(this));
        this.connection.onSignatureHelp(this.onSignatureHelp.bind(this));
        this.connection.onDocumentSymbol(this.onDocumentSymbol.bind(this));
        this.connection.onDefinition(this.onDefinition.bind(this));
        this.connection.onFoldingRanges(this.onFoldingRange.bind(this));
        this.documentService.onChange(this.onChange.bind(this));
        this.connection.listen();
    }
    /**
     * Initialize the server from the client connection
     * @param params initialization parameters
     */
    onInitialize(params) {
        const { capabilities, rootPath, rootUri } = params;
        this.connection.console.log(`[KLS Server(${process.pid})] Started and initialize received.`);
        // does the client support configurations
        this.configuration.clientCapability.hasConfiguration = !!(capabilities.workspace && !!capabilities.workspace.configuration);
        // does the client support workspace folders
        this.configuration.clientCapability.hasWorkspaceFolder = !!(capabilities.workspace && !!capabilities.workspace.workspaceFolders);
        // get root path if it exists
        if (rootPath) {
            this.configuration.workspaceFolder = rootPath;
        }
        // get root uri if it exists
        if (rootUri) {
            this.setUri(rootUri);
            this.configuration.workspaceUri = rootUri;
        }
        return {
            capabilities: {
                textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Incremental,
                // Tell the client that the server supports code completion
                completionProvider: {
                    resolveProvider: true,
                    triggerCharacters: [':', '(', ', '],
                },
                // Tell the client that the server support signature help
                signatureHelpProvider: {
                    triggerCharacters: ['(', ',', ', '],
                },
                // indicate other capabilities
                renameProvider: true,
                documentHighlightProvider: true,
                hoverProvider: true,
                referencesProvider: true,
                documentSymbolProvider: true,
                definitionProvider: true,
                foldingRangeProvider: true,
            },
        };
    }
    /**
     * Post initialization register additional hooks and retrieve client
     * configurations
     * @param _ initialized parameters
     */
    onInitialized(_) {
        return __awaiter(this, void 0, void 0, function* () {
            const { clientCapability } = this.configuration;
            // register for all configuration changes.
            if (clientCapability.hasConfiguration) {
                this.connection.client.register(vscode_languageserver_1.DidChangeConfigurationNotification.type, {
                    section: constants_1.serverName,
                });
            }
            // register workspace changes
            if (clientCapability.hasWorkspaceFolder) {
                this.connection.workspace.onDidChangeWorkspaceFolders(_ => {
                    // TODO dump all documents
                    this.logger.log('Workspace folder change event received.');
                });
            }
            const clientConfig = yield this.getDocumentSettings();
            this.updateServer(clientConfig);
        });
    }
    /**
     * Update the server configuration when the client signals a change in it's configuration for
     * the kos-language-server
     * @param change The updated settings
     */
    onDidChangeConfiguration(change) {
        const { clientCapability } = this.configuration;
        if (clientCapability.hasConfiguration) {
            if (change.settings && constants_1.serverName in change.settings) {
                Object.assign(this.configuration.clientConfig, serverUtils_1.defaultClientConfiguration, change.settings[constants_1.serverName]);
            }
            // update server on client config
            this.updateServer(this.configuration.clientConfig);
        }
    }
    /**
     * Respond to completion requests from the client. This handler currently provides
     * both symbol completion as well as suffix completion.
     * @param completion the parameters describing the completion request
     */
    onCompletion(completion) {
        return __awaiter(this, void 0, void 0, function* () {
            const { context } = completion;
            try {
                // check if suffix completion
                if (!typeGuards_1.empty(context) && !typeGuards_1.empty(context.triggerCharacter)) {
                    const { triggerCharacter } = context;
                    if (triggerCharacter === ':') {
                        return serverUtils_1.suffixCompletionItems(this, completion);
                    }
                }
                // complete base symbols
                return yield serverUtils_1.symbolCompletionItems(this, completion, this.configuration.keywords);
                // catch any errors
            }
            catch (err) {
                logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
                return undefined;
            }
        });
    }
    /**
     * This handler provider completion item resolution capability. This provides
     * additional information for the currently completion item selection
     * @param completionItem the item to resolve further
     */
    onCompletionResolve(completionItem) {
        try {
            const token = completionItem.data;
            if (!typeGuards_1.empty(token)) {
            }
            return completionItem;
        }
        catch (err) {
            logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
            return completionItem;
        }
    }
    /**
     * This handler provider rename capabilities. This allows a client to highlight
     * as symbol and provide a new name that will change for all known symbols
     * @param rename information describing what and where a rename should occur
     */
    onRenameRequest(rename) {
        return __awaiter(this, void 0, void 0, function* () {
            const { newName, position, textDocument } = rename;
            const scanner = new scanner_1.Scanner(newName);
            const { tokens, scanErrors } = scanner.scanTokens();
            // check if rename is valid
            if (scanErrors.length > 0 ||
                tokens.length !== 1 ||
                !tokentypes_1.isValidIdentifier(tokens[0].type)) {
                return undefined;
            }
            const locations = yield this.getUsageLocations(position, textDocument.uri);
            if (typeGuards_1.empty(locations)) {
                return undefined;
            }
            const changes = {};
            for (const location of locations) {
                if (!changes.hasOwnProperty(location.uri)) {
                    changes[location.uri] = [];
                }
                changes[location.uri].push(vscode_languageserver_1.TextEdit.replace(location.range, newName));
            }
            return { changes };
        });
    }
    /**
     * This handler provides highlight within a requested document. This allows the client
     * to highlight and symbol and other instances of that symbol to also highlight.
     * @param positionParams the position of the highlight request
     */
    onDocumentHighlight(positionParams) {
        return __awaiter(this, void 0, void 0, function* () {
            const { position } = positionParams;
            const { uri } = positionParams.textDocument;
            const locations = yield this.getFileUsageRanges(position, uri);
            return typeGuards_1.empty(locations)
                ? []
                : locations.map(range => ({ range: clean_1.cleanRange(range) }));
        });
    }
    /**
     * This handler provides on hover capability for symbols in a document. This allows additional
     * information to be displayed to the user about symbols throughout the document
     * @param positionParams the position of the hover request
     */
    onHover(positionParams) {
        return __awaiter(this, void 0, void 0, function* () {
            const { position } = positionParams;
            const { uri } = positionParams.textDocument;
            const token = yield this.getToken(position, uri);
            if (typeGuards_1.empty(token)) {
                return undefined;
            }
            const type = typeUtilities_1.tokenTrackedType(token);
            const { tracker } = token;
            let label;
            let symbolKind;
            if (!typeGuards_1.empty(tracker)) {
                symbolKind = types_1.KsSymbolKind[tracker.declared.symbol.tag];
                const { name } = tracker.declared.symbol;
                label = typeof name === 'string' ? name : name.lexeme;
            }
            else {
                symbolKind = 'literal';
                label = token.lexeme;
            }
            if (typeGuards_1.empty(type)) {
                return undefined;
            }
            return {
                contents: {
                    // Note doesn't does do much other than format it as code
                    // may look into adding type def syntax highlighting
                    language: 'kos',
                    value: `(${symbolKind}) ${label}: ${type.toString()} `,
                },
                range: {
                    start: clean_1.cleanPosition(token.start),
                    end: clean_1.cleanPosition(token.end),
                },
            };
        });
    }
    /**
     * This handler provides reference capabilities to symbols in a document. This allows a client
     * to identify all positions that a symbol is used in the document or attached documents
     * @param reference parameters describing the reference request
     */
    onReference(reference) {
        return __awaiter(this, void 0, void 0, function* () {
            const { position } = reference;
            const { uri } = reference.textDocument;
            const locations = yield this.getUsageLocations(position, uri);
            return locations && locations.map(loc => clean_1.cleanLocation(loc));
        });
    }
    /**
     * This handler provides signature help suffixes and function within the document. This
     * provides extra context to the client such as the current parameter
     * @param positionParams the position of the signature request
     */
    onSignatureHelp(positionParams) {
        return __awaiter(this, void 0, void 0, function* () {
            const { position } = positionParams;
            const { uri } = positionParams.textDocument;
            const result = yield this.getFunctionAtPosition(position, uri);
            if (typeGuards_1.empty(result))
                return serverUtils_1.defaultSignature();
            const { tracker, index } = result;
            let label = typeof tracker.declared.symbol.name === 'string'
                ? tracker.declared.symbol.name
                : tracker.declared.symbol.name.lexeme;
            const type = tracker.getType({
                uri,
                range: { start: position, end: position },
            });
            if (typeGuards_1.empty(type)) {
                return serverUtils_1.defaultSignature();
            }
            switch (type.kind) {
                case 4 /* function */:
                case 2 /* suffix */:
                    const callSignature = type.getCallSignature();
                    if (typeGuards_1.empty(callSignature)) {
                        return serverUtils_1.defaultSignature();
                    }
                    let start = label.length + 1;
                    const params = callSignature.params();
                    const paramInfos = [];
                    // check if normal or variadic type
                    if (params.length === 1 && params[0].kind === 1 /* variadic */) {
                        // generate variadic labels
                        const variadicLabel = params[0].toString();
                        paramInfos.push(vscode_languageserver_1.ParameterInformation.create([start, start + variadicLabel.length]));
                        label = `${label}(${variadicLabel})`;
                    }
                    else if (params.length > 0) {
                        // generate normal labels
                        const labels = [];
                        for (let i = 0; i < params.length - 1; i += 1) {
                            const paramLabel = `${params[i].toString()}, `;
                            paramInfos.push(vscode_languageserver_1.ParameterInformation.create([
                                start,
                                start + paramLabel.length - 2,
                            ]));
                            labels.push(paramLabel);
                            start = start + paramLabel.length;
                        }
                        const paramLabel = `${params[params.length - 1].toString()}`;
                        paramInfos.push(vscode_languageserver_1.ParameterInformation.create([start, start + paramLabel.length]));
                        labels.push(paramLabel);
                        label = `${label}(${labels.join('')})`;
                    }
                    return {
                        signatures: [
                            vscode_languageserver_1.SignatureInformation.create(label, undefined, ...paramInfos),
                        ],
                        activeParameter: index,
                        activeSignature: null,
                    };
                default:
                    return serverUtils_1.defaultSignature();
            }
        });
    }
    /**
     * This handler provides document symbol capabilities. This provides a list of all
     * symbols that are located within a given document
     * @param documentSymbol the document to provide symbols for
     */
    onDocumentSymbol(documentSymbol) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield serverUtils_1.documentSymbols(this, documentSymbol);
        });
    }
    /**
     * This handler provides go to definition capabilities. When a client requests a symbol
     * go to definition this provides the location if it exists
     * @param positionParams the position of the definition request
     */
    onDefinition(positionParams) {
        return __awaiter(this, void 0, void 0, function* () {
            const { position } = positionParams;
            const { uri } = positionParams.textDocument;
            const location = yield this.getDeclarationLocation(position, uri);
            return location && clean_1.cleanLocation(location);
        });
    }
    /**
     * This handler provide folding region capabilities. The client will ask for available folding
     * region in which this will respond with the ranges defined by #region and #endregion
     * @param foldingParams the document to preform folding analysis on
     */
    onFoldingRange(foldingParams) {
        return __awaiter(this, void 0, void 0, function* () {
            const { uri } = foldingParams.textDocument;
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (typeGuards_1.empty(documentInfo)) {
                return [];
            }
            const { script, regions } = documentInfo;
            return this.foldableService.findRegions(script, regions);
        });
    }
    /**
     * Respond to updates made to document by the client. This method
     * will parse and update the internal state of affects scripts
     * reporting errors to the client as they are discovered
     * @param document the updated document
     */
    onChange(document) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const diagnosticResults = yield this.analysisService.validateDocument(document.uri, document.text);
                const total = diagnosticResults.length;
                const diagnosticMap = new Map();
                // retrieve diagnostics from analyzer
                for (const diagnostic of diagnosticResults) {
                    const uriDiagnostics = diagnosticMap.get(diagnostic.uri);
                    if (typeGuards_1.empty(uriDiagnostics)) {
                        diagnosticMap.set(diagnostic.uri, [clean_1.cleanDiagnostic(diagnostic)]);
                    }
                    else {
                        uriDiagnostics.push(clean_1.cleanDiagnostic(diagnostic));
                    }
                }
                // send diagnostics to each document reported
                for (const [uri, diagnostics] of diagnosticMap.entries()) {
                    this.connection.sendDiagnostics({
                        uri,
                        diagnostics,
                    });
                }
                // if not problems found clear out diagnostics
                if (total === 0) {
                    this.connection.sendDiagnostics({
                        uri: document.uri,
                        diagnostics: [],
                    });
                }
            }
            catch (err) {
                // report any exceptions to the client
                logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
            }
        });
    }
    /**
     * Get document settings from the client. If the client does not support
     * have configurations then return the default configurations.
     */
    getDocumentSettings() {
        if (!this.configuration.clientCapability.hasConfiguration) {
            return Promise.resolve(serverUtils_1.defaultClientConfiguration);
        }
        return this.connection.workspace.getConfiguration({
            scopeUri: this.workspaceUri,
            section: constants_1.serverName,
        });
    }
    /**
     * Update the servers configuration in response to a change in the client configuration
     * @param clientConfig client configuration
     */
    updateServer(clientConfig) {
        this.configuration.clientConfig = clientConfig;
        const casePreference = serverUtils_1.caseMapper(clientConfig.completionCase);
        const logPreference = serverUtils_1.logMapper(clientConfig.trace.server.level);
        this.setCase(casePreference);
        this.logger.level = logPreference;
    }
    /**
     * Set the volume 0 path for the analyzer
     * @param uri path of volume 0
     */
    setUri(uri) {
        const parsed = vscode_uri_1.URI.parse(uri);
        this.documentService.setVolume0Uri(parsed);
        this.workspaceUri = uri;
    }
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind) {
        this.configuration.keywords = constants_1.keywordCompletions(caseKind);
        this.analysisService.setCase(caseKind);
    }
    /**
     * Get the token at the provided position in the text document
     * @param pos position in the text document
     * @param uri uri of the text document
     */
    getToken(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (typeGuards_1.empty(documentInfo)) {
                return undefined;
            }
            // try to find an symbol at the position
            const { script } = documentInfo;
            const finder = new scriptFind_1.ScriptFind();
            const result = finder.find(script, pos);
            return result && result.token;
        });
    }
    /**
     * Get the declaration location for the token at the provided position
     * @param pos position in the document
     * @param uri uri of the document
     */
    getDeclarationLocation(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (typeGuards_1.empty(documentInfo)) {
                return undefined;
            }
            // try to find an symbol at the position
            const { script } = documentInfo;
            const finder = new scriptFind_1.ScriptFind();
            const result = finder.find(script, pos, Stmt.Run, Stmt.RunPath, Stmt.RunOncePath);
            if (typeGuards_1.empty(result)) {
                return undefined;
            }
            // check if symbols exists
            const { token, node } = result;
            if (typeGuards_1.empty(token.tracker)) {
                // if no tracker it might be a run statment
                if (node instanceof Stmt.Run ||
                    node instanceof Stmt.RunPath ||
                    node instanceof Stmt.RunOncePath) {
                    const pathResolver = new pathResolver_1.PathResolver(this.configuration.workspaceUri);
                    // get the kos run path
                    const kosPath = pathResolver_1.runPath(node);
                    if (typeof kosPath !== 'string') {
                        return undefined;
                    }
                    // resolve to the file system
                    const path = pathResolver.resolveUri(node.toLocation(uri), kosPath);
                    if (typeGuards_1.empty(path)) {
                        return undefined;
                    }
                    // check if file exists then
                    if (fs_1.existsSync(path.fsPath)) {
                        return vscode_languageserver_1.Location.create(path.toString(), vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(0, 0), vscode_languageserver_1.Position.create(0, 0)));
                    }
                    // if we didn't find try to normalize the path with extension .ks
                    const result = pathResolver_1.normalizeExtensions(path);
                    if (typeGuards_1.empty(result)) {
                        return undefined;
                    }
                    const normalized = vscode_uri_1.URI.parse(result);
                    if (!fs_1.existsSync(normalized.fsPath)) {
                        return undefined;
                    }
                    return vscode_languageserver_1.Location.create(normalized.toString(), vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(0, 0), vscode_languageserver_1.Position.create(0, 0)));
                }
                return undefined;
            }
            const { declared } = token.tracker;
            // exit if undefined
            if (declared.uri === constants_1.builtIn) {
                return undefined;
            }
            return typeof declared.symbol.name !== 'string'
                ? declared.symbol.name
                : undefined;
        });
    }
    /**
     * Get all usage locations in all files
     * @param pos position in document
     * @param uri uri of document
     */
    getUsageLocations(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (typeGuards_1.empty(documentInfo) ||
                typeGuards_1.empty(documentInfo.symbolTable) ||
                typeGuards_1.empty(documentInfo.script)) {
                return undefined;
            }
            // try to find the symbol at the position
            const { symbolTable: symbolsTable, script } = documentInfo;
            const finder = new scriptFind_1.ScriptFind();
            const result = finder.find(script, pos);
            if (typeGuards_1.empty(result)) {
                return undefined;
            }
            // try to find the tracker at a given position
            const { token } = result;
            const tracker = symbolsTable.scopedNamedTracker(pos, token.lookup);
            if (typeGuards_1.empty(tracker)) {
                return undefined;
            }
            return tracker.usages
                .map(usage => usage)
                .concat(tracker.declared.symbol.name)
                .filter(location => location.uri !== constants_1.builtIn);
        });
    }
    /**
     * Get all usage ranges in a provide file
     * @param pos position in document
     * @param uri uri of document
     */
    getFileUsageRanges(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const locations = yield this.getUsageLocations(pos, uri);
            if (typeGuards_1.empty(locations)) {
                return locations;
            }
            return locations.filter(loc => loc.uri === uri).map(loc => loc.range);
        });
    }
    /**
     * Get all symbols in scope at a particular location in the file
     * @param pos position in document
     * @param uri document uri
     */
    getScopedSymbols(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (!typeGuards_1.empty(documentInfo) && !typeGuards_1.empty(documentInfo.symbolTable)) {
                return documentInfo.symbolTable.scopedSymbols(pos);
            }
            return [];
        });
    }
    /**
     * Get all symbols in a provided file
     * @param uri document uri
     */
    getAllFileSymbols(uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (!typeGuards_1.empty(documentInfo) && !typeGuards_1.empty(documentInfo.symbolTable)) {
                return documentInfo.symbolTable.allSymbols();
            }
            return [];
        });
    }
    /**
     * Get a function located at the current location
     * @param pos position in the document
     * @param uri document uri
     */
    getFunctionAtPosition(pos, uri) {
        return __awaiter(this, void 0, void 0, function* () {
            // we need the document info to lookup a signature
            const documentInfo = yield this.analysisService.getInfo(uri);
            if (typeGuards_1.empty(documentInfo))
                return undefined;
            const { script } = documentInfo;
            const finder = new scriptFind_1.ScriptFind();
            // attempt to find a token here get surround invalid Stmt context
            const outerResult = finder.find(script, pos, Expr.Suffix, Stmt.Invalid);
            const innerResult = finder.find(script, pos, SuffixTerm.Call);
            let index = 0;
            if (!typeGuards_1.empty(innerResult) &&
                !typeGuards_1.empty(innerResult.node) &&
                innerResult.node instanceof SuffixTerm.Call) {
                index =
                    innerResult.node.args.length > 0 ? innerResult.node.args.length - 1 : 0;
            }
            // currently we only support invalid statements for signature completion
            // we could possible support call expressions as well
            if (typeGuards_1.empty(outerResult) || typeGuards_1.empty(outerResult.node)) {
                return undefined;
            }
            // determine the identifier of the invalid statement and parameter index
            const { node } = outerResult;
            // check if suffix
            if (node instanceof Expr.Suffix) {
                const tracker = node.mostResolveTracker();
                if (typeGuards_1.empty(tracker)) {
                    return undefined;
                }
                switch (tracker.declared.symbol.tag) {
                    case types_1.KsSymbolKind.function:
                    case types_1.KsSymbolKind.suffix:
                        return {
                            index,
                            tracker,
                        };
                    default:
                        return undefined;
                }
            }
            // check if invalid statement
            if (node instanceof Stmt.Invalid) {
                const { ranges } = node;
                const indices = positionUtils_1.binarySearchIndex(ranges, pos);
                const start = Array.isArray(indices) ? indices[0] : indices;
                for (let i = start; i >= 0; i -= 1) {
                    const element = ranges[i];
                    if (element instanceof Expr.Suffix) {
                        const tracker = element.mostResolveTracker();
                        if (!typeGuards_1.empty(tracker)) {
                            switch (tracker.declared.symbol.tag) {
                                case types_1.KsSymbolKind.function:
                                case types_1.KsSymbolKind.suffix:
                                    return {
                                        index,
                                        tracker,
                                    };
                                default:
                                    return undefined;
                            }
                        }
                    }
                }
            }
            return undefined;
        });
    }
}
exports.KLS = KLS;
//# sourceMappingURL=kls.js.map