"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
class TreeTraverse {
    constructor() { }
    // find an statement
    stmtAction(stmt) {
        return stmt.accept(this);
    }
    // find an expression
    exprAction(expr) {
        return expr.accept(this);
    }
    // find an expression
    suffixTermAction(suffixTerm) {
        return suffixTerm.accept(this);
    }
    visitDeclVariable(decl) {
        this.exprAction(decl.value);
    }
    visitDeclLock(decl) {
        this.exprAction(decl.value);
    }
    visitDeclFunction(decl) {
        this.stmtAction(decl.block);
    }
    visitDeclParameter(decl) {
        for (const parameter of decl.optionalParameters) {
            this.exprAction(parameter.value);
        }
    }
    visitStmtInvalid(_) { }
    visitBlock(stmt) {
        for (const childStmt of stmt.stmts) {
            this.stmtAction(childStmt);
        }
    }
    visitExpr(stmt) {
        this.exprAction(stmt.suffix);
    }
    visitOnOff(stmt) {
        this.exprAction(stmt.suffix);
    }
    visitCommand(_) { }
    visitCommandExpr(stmt) {
        this.exprAction(stmt.expr);
    }
    visitUnset(_) { }
    visitUnlock(_) { }
    visitSet(stmt) {
        this.exprAction(stmt.suffix);
        this.exprAction(stmt.value);
    }
    visitLazyGlobal(_) { }
    visitIf(stmt) {
        this.exprAction(stmt.condition);
        this.stmtAction(stmt.body);
        if (!typeGuards_1.empty(stmt.elseStmt)) {
            this.stmtAction(stmt.elseStmt);
        }
    }
    visitElse(stmt) {
        this.stmtAction(stmt.body);
    }
    visitUntil(stmt) {
        this.exprAction(stmt.condition);
        this.stmtAction(stmt.body);
    }
    visitFrom(stmt) {
        this.stmtAction(stmt.initializer);
        this.exprAction(stmt.condition);
        this.stmtAction(stmt.increment);
        this.stmtAction(stmt.body);
    }
    visitWhen(stmt) {
        this.exprAction(stmt.condition);
        this.stmtAction(stmt.body);
    }
    visitReturn(stmt) {
        if (!typeGuards_1.empty(stmt.value)) {
            this.exprAction(stmt.value);
        }
    }
    visitBreak(_) { }
    visitSwitch(stmt) {
        this.exprAction(stmt.target);
    }
    visitFor(stmt) {
        this.exprAction(stmt.collection);
        this.stmtAction(stmt.body);
    }
    visitOn(stmt) {
        this.exprAction(stmt.suffix);
        this.stmtAction(stmt.body);
    }
    visitToggle(stmt) {
        this.exprAction(stmt.suffix);
    }
    visitWait(stmt) {
        this.exprAction(stmt.expr);
    }
    visitLog(stmt) {
        this.exprAction(stmt.expr);
        this.exprAction(stmt.target);
    }
    visitCopy(stmt) {
        this.exprAction(stmt.target);
        this.exprAction(stmt.destination);
    }
    visitRename(stmt) {
        this.exprAction(stmt.target);
        this.exprAction(stmt.alternative);
    }
    visitDelete(stmt) {
        this.exprAction(stmt.target);
        if (!typeGuards_1.empty(stmt.volume)) {
            this.exprAction(stmt.volume);
        }
    }
    visitRun(stmt) {
        if (!typeGuards_1.empty(stmt.args)) {
            for (const arg of stmt.args) {
                this.exprAction(arg);
            }
        }
        if (!typeGuards_1.empty(stmt.expr)) {
            this.exprAction(stmt.expr);
        }
    }
    visitRunPath(stmt) {
        this.exprAction(stmt.expr);
        if (!typeGuards_1.empty(stmt.args)) {
            for (const arg of stmt.args) {
                this.exprAction(arg);
            }
        }
    }
    visitRunPathOnce(stmt) {
        this.exprAction(stmt.expr);
        if (!typeGuards_1.empty(stmt.args)) {
            for (const arg of stmt.args) {
                this.exprAction(arg);
            }
        }
    }
    visitCompile(stmt) {
        this.exprAction(stmt.target);
        if (!typeGuards_1.empty(stmt.destination)) {
            this.exprAction(stmt.destination);
        }
    }
    visitList(_) { }
    visitEmpty(_) { }
    visitPrint(stmt) {
        this.exprAction(stmt.expr);
        if (!typeGuards_1.empty(stmt.x)) {
            this.exprAction(stmt.x);
        }
        if (!typeGuards_1.empty(stmt.y)) {
            this.exprAction(stmt.y);
        }
    }
    visitExprInvalid(_) { }
    visitTernary(expr) {
        this.exprAction(expr.condition);
        this.exprAction(expr.trueBranch);
        this.exprAction(expr.falseBranch);
    }
    visitBinary(expr) {
        this.exprAction(expr.left);
        this.exprAction(expr.right);
    }
    visitUnary(expr) {
        this.exprAction(expr.factor);
    }
    visitFactor(expr) {
        this.exprAction(expr.suffix);
        this.exprAction(expr.exponent);
    }
    visitSuffix(expr) {
        this.suffixTermAction(expr.suffixTerm);
        if (!typeGuards_1.empty(expr.trailer)) {
            this.suffixTermAction(expr.trailer);
        }
    }
    visitLambda(expr) {
        this.stmtAction(expr.block);
    }
    visitSuffixTermInvalid(_) { }
    visitSuffixTrailer(suffixTerm) {
        this.suffixTermAction(suffixTerm.suffixTerm);
        if (!typeGuards_1.empty(suffixTerm.trailer)) {
            this.suffixTermAction(suffixTerm.trailer);
        }
    }
    visitSuffixTerm(suffixTerm) {
        this.suffixTermAction(suffixTerm.atom);
        for (const trailer of suffixTerm.trailers) {
            this.suffixTermAction(trailer);
        }
    }
    visitCall(suffixTerm) {
        for (const arg of suffixTerm.args) {
            this.exprAction(arg);
        }
    }
    visitArrayIndex(_) { }
    visitArrayBracket(suffixTerm) {
        this.exprAction(suffixTerm.index);
    }
    visitDelegate(_) { }
    visitLiteral(_) { }
    visitIdentifier(_) { }
    visitGrouping(suffixTerm) {
        this.exprAction(suffixTerm.expr);
    }
}
exports.TreeTraverse = TreeTraverse;
//# sourceMappingURL=treeTraverse.js.map