import { DiagnosticUri, IDocumentInfo as DocumentInfo } from '../types';
import { DocumentService } from './documentService';
export declare class AnalysisService {
    /**
     * A logger for message and error logger
     */
    private readonly logger;
    /**
     * A tracer to follow where exceptions occur
     */
    private readonly tracer;
    /**
     * Performance observer for tracking analysis speed
     */
    private readonly observer;
    /**
     * A service to load documents
     */
    private readonly documentService;
    /**
     * Document information
     */
    private readonly documentInfos;
    /**
     * The current loaded standard library
     */
    private standardLibrary;
    /**
     * The current loaded celestial body library
     */
    private bodyLibrary;
    /**
     * The current requested completion case
     */
    private caseKind;
    /**
     * Construct a new analysis service
     * @param caseKind the typing case of the build in standard and body libraries
     * @param logger A logger to log performance and exception
     * @param tracer a tracer to location exceptions
     * @param documentService The document service to load new files from disk
     */
    constructor(caseKind: CaseKind, logger: ILogger, tracer: ITracer, documentService: DocumentService);
    /**
     * Validate a document in asynchronous stages. This produces diagnostics about known errors or
     * potential problems in the provided script
     * @param uri uri of the document
     * @param text source text of the document
     */
    validateDocument(uri: string, text: string): Promise<DiagnosticUri[]>;
    /**
     * Get a document info if it exists
     */
    getInfo(uri: string): Promise<Maybe<DocumentInfo>>;
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind: CaseKind): void;
    /**
     * Main validation function for a document. Lexically and semantically understands a document.
     * Will additionally perform the same analysis on other run scripts found in this script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    private validateDocument_;
    /**
     * Perform lexical analysis on the provided provided source text
     * @param uri uri to document
     * @param text source text of document
     */
    private lexicalAnalysisDocument;
    /**
     * Perform semantic analysis on the an ast and provided dependency symbol tables
     * @param uri uri to document
     * @param script Ast of the document
     * @param tables symbol tables that are dependencies of this document
     */
    private semanticAnalysisDocument;
    /**
     * Load dependencies either from cache or by perform analysis on them as well
     * @param uri uri to document
     * @param script Ast of the document
     * @param depth current resolving depth TODO actually check cycles
     */
    private loadDependencies;
    /**
     * Get all valid uris from the documents run statements
     * @param uri uri of the calling document
     * @param runStmts run statements in the document
     */
    private loadDocuments;
    /**
     * Get the symbol table corresponding to the active standard library. For now this
     * will just be the default library but this may allow us to eventually support multiple
     * version of kOS
     */
    private activeStandardLibrary;
    /**
     * Get the symbol table corresponding active set of celestial bodies for the user.
     * This allows for bodies other than that in stock ksp to be incorporated
     */
    private activeBodyLibrary;
}
