"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const symbolTableBuilder_1 = require("../analysis/symbolTableBuilder");
const preResolver_1 = require("../analysis/preResolver");
const perf_hooks_1 = require("perf_hooks");
const typeChecker_1 = require("../typeChecker/typeChecker");
const typeGuards_1 = require("../utilities/typeGuards");
const scanner_1 = require("../scanner/scanner");
const parser_1 = require("../parser/parser");
const vscode_languageserver_1 = require("vscode-languageserver");
const serverUtils_1 = require("../utilities/serverUtils");
const logger_1 = require("../utilities/logger");
const resolver_1 = require("../analysis/resolver");
const pathResolver_1 = require("../utilities/pathResolver");
const standardLibrary_1 = require("../analysis/standardLibrary");
class AnalysisService {
    /**
     * Construct a new analysis service
     * @param caseKind the typing case of the build in standard and body libraries
     * @param logger A logger to log performance and exception
     * @param tracer a tracer to location exceptions
     * @param documentService The document service to load new files from disk
     */
    constructor(caseKind, logger, tracer, documentService) {
        this.logger = logger;
        this.tracer = tracer;
        this.documentService = documentService;
        this.documentInfos = new Map();
        this.caseKind = caseKind;
        this.standardLibrary = standardLibrary_1.standardLibraryBuilder(caseKind);
        this.bodyLibrary = standardLibrary_1.bodyLibraryBuilder(caseKind);
        this.observer = new perf_hooks_1.PerformanceObserver(list => {
            this.logger.info('');
            this.logger.info('-------- Performance ---------');
            for (const entry of list.getEntries()) {
                this.logger.info(`${entry.name} took ${entry.duration} ms`);
            }
            this.logger.info('------------------------------');
        });
        this.observer.observe({ entryTypes: ['measure'], buffered: true });
    }
    /**
     * Validate a document in asynchronous stages. This produces diagnostics about known errors or
     * potential problems in the provided script
     * @param uri uri of the document
     * @param text source text of the document
     */
    validateDocument(uri, text) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const result = yield this.validateDocument_(uri, text, 0);
                if (!typeGuards_1.empty(result)) {
                    this.documentInfos.set(uri, result);
                }
                return typeGuards_1.empty(result) ? [] : result.diagnostics;
            }
            catch (err) {
                logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
                return [];
            }
        });
    }
    /**
     * Get a document info if it exists
     */
    getInfo(uri) {
        return __awaiter(this, void 0, void 0, function* () {
            // if we already have the document loaded return it
            const documentInfo = this.documentInfos.get(uri);
            if (!typeGuards_1.empty(documentInfo)) {
                return documentInfo;
            }
            try {
                const document = yield this.documentService.loadDocument(uri);
                if (typeGuards_1.empty(document)) {
                    return undefined;
                }
                const result = yield this.validateDocument_(uri, document.getText(), 0);
                if (!typeGuards_1.empty(result)) {
                    this.documentInfos.set(uri, result);
                }
                return result;
            }
            catch (err) {
                logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
                return undefined;
            }
        });
    }
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind) {
        if (this.caseKind !== caseKind) {
            this.caseKind = caseKind;
            this.standardLibrary = standardLibrary_1.standardLibraryBuilder(caseKind);
            this.bodyLibrary = standardLibrary_1.bodyLibraryBuilder(caseKind);
        }
    }
    /**
     * Main validation function for a document. Lexically and semantically understands a document.
     * Will additionally perform the same analysis on other run scripts found in this script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    validateDocument_(uri, text, depth) {
        return __awaiter(this, void 0, void 0, function* () {
            if (depth > 10) {
                return undefined;
            }
            // preform lexical analysis
            const { script, regions, scannerDiagnostics, parserDiagnostics, } = this.lexicalAnalysisDocument(uri, text);
            // load dependencies found in the ast
            const { documentInfos, loadDiagnostics } = yield this.loadDependencies(uri, script, depth);
            // add standard library to dependencies
            const dependencyTables = new Set([
                this.activeStandardLibrary(),
                this.activeBodyLibrary(),
            ]);
            // add run statement dependencies and their dependencies
            for (const documentInfo of documentInfos) {
                dependencyTables.add(documentInfo.symbolTable);
            }
            // perform semantic analysis
            const { resolverDiagnostics, typeDiagnostics, symbolTable, } = this.semanticAnalysisDocument(uri, script, dependencyTables);
            // clear performance observer marks
            perf_hooks_1.performance.clearMarks();
            // generate the document info
            return {
                script,
                regions,
                symbolTable,
                diagnostics: [
                    ...scannerDiagnostics,
                    ...parserDiagnostics,
                    ...loadDiagnostics,
                    ...resolverDiagnostics,
                    ...typeDiagnostics,
                ],
            };
        });
    }
    /**
     * Perform lexical analysis on the provided provided source text
     * @param uri uri to document
     * @param text source text of document
     */
    lexicalAnalysisDocument(uri, text) {
        this.logger.verbose('');
        this.logger.verbose('-------------Lexical Analysis------------');
        perf_hooks_1.performance.mark('scanner-start');
        const scanner = new scanner_1.Scanner(text, uri, this.logger, this.tracer);
        const { tokens, scanErrors, regions } = scanner.scanTokens();
        perf_hooks_1.performance.mark('scanner-end');
        // if scanner found errors report those immediately
        if (scanErrors.length > 0) {
            this.logger.warn(`Scanning encountered ${scanErrors.length} Errors.`);
        }
        perf_hooks_1.performance.mark('parser-start');
        const parser = new parser_1.Parser(uri, tokens, this.logger, this.tracer);
        const { script, parseErrors } = parser.parse();
        perf_hooks_1.performance.mark('parser-end');
        // measure performance
        perf_hooks_1.performance.measure('Scanner', 'scanner-start', 'scanner-end');
        perf_hooks_1.performance.measure('Parser', 'parser-start', 'parser-end');
        perf_hooks_1.performance.clearMarks();
        this.logger.verbose('--------------------------------------');
        // generate lexical diagnostics
        const scannerDiagnostics = scanErrors.map(scanError => serverUtils_1.addDiagnosticsUri(scanError, uri));
        const parserDiagnostics = parseErrors.length === 0
            ? []
            : parseErrors
                .map(error => error.inner.concat(error))
                .reduce((acc, current) => acc.concat(current))
                .map(error => serverUtils_1.parseToDiagnostics(error, uri));
        return {
            scannerDiagnostics,
            parserDiagnostics,
            script,
            regions,
        };
    }
    /**
     * Perform semantic analysis on the an ast and provided dependency symbol tables
     * @param uri uri to document
     * @param script Ast of the document
     * @param tables symbol tables that are dependencies of this document
     */
    semanticAnalysisDocument(uri, script, tables) {
        this.logger.verbose('');
        this.logger.verbose('-------------Semantic Analysis------------');
        // generate a scope manager for resolving
        const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder(uri, this.logger);
        let oldDocumentInfo = this.documentInfos.get(uri);
        // add symbol tables that are dependent
        if (!typeGuards_1.empty(oldDocumentInfo)) {
            for (const dependent of oldDocumentInfo.symbolTable.dependentTables) {
                symbolTableBuilder.linkDependent(dependent);
            }
            oldDocumentInfo.symbolTable.removeSelf();
        }
        // add symbol tables that are dependencies
        for (const symbolTable of tables) {
            symbolTableBuilder.linkDependency(symbolTable);
        }
        // generate resolvers
        const preResolver = new preResolver_1.PreResolver(script, symbolTableBuilder, this.logger, this.tracer);
        const resolver = new resolver_1.Resolver(script, symbolTableBuilder, this.logger, this.tracer);
        // traverse the ast to find functions to pre populate symbol table
        perf_hooks_1.performance.mark('pre-resolver-start');
        const resolverDiagnostics = preResolver
            .resolve()
            .map(error => serverUtils_1.addDiagnosticsUri(error, uri));
        perf_hooks_1.performance.mark('pre-resolver-end');
        // traverse the ast again to resolve the remaining symbols
        perf_hooks_1.performance.mark('resolver-start');
        resolverDiagnostics.push(...resolver.resolve().map(error => serverUtils_1.addDiagnosticsUri(error, uri)));
        // find scopes were symbols were never used
        resolverDiagnostics.push(...symbolTableBuilder
            .findUnused()
            .map(error => serverUtils_1.addDiagnosticsUri(error, uri)));
        perf_hooks_1.performance.mark('resolver-end');
        // build the final symbol table
        const symbolTable = symbolTableBuilder.build();
        // make sure to delete references so scope manager can be gc'ed
        if (!typeGuards_1.empty(oldDocumentInfo)) {
            oldDocumentInfo.symbolTable.removeSelf();
            oldDocumentInfo = undefined;
        }
        // perform type checking
        const typeChecker = new typeChecker_1.TypeChecker(script, this.logger, this.tracer);
        perf_hooks_1.performance.mark('type-checking-start');
        const typeDiagnostics = typeChecker
            .check()
            .map(error => serverUtils_1.addDiagnosticsUri(error, uri));
        perf_hooks_1.performance.mark('type-checking-end');
        // measure performance
        perf_hooks_1.performance.measure('Pre Resolver', 'pre-resolver-start', 'pre-resolver-end');
        perf_hooks_1.performance.measure('Resolver', 'resolver-start', 'resolver-end');
        perf_hooks_1.performance.measure('Type Checking', 'type-checking-start', 'type-checking-end');
        this.logger.verbose('--------------------------------------');
        return {
            symbolTable,
            typeDiagnostics,
            resolverDiagnostics,
        };
    }
    /**
     * Load dependencies either from cache or by perform analysis on them as well
     * @param uri uri to document
     * @param script Ast of the document
     * @param depth current resolving depth TODO actually check cycles
     */
    loadDependencies(uri, script, depth) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = {
                documentInfos: [],
                loadDiagnostics: [],
            };
            // if any run statement exist get uri then load
            if (script.runStmts.length > 0 && this.documentService.ready) {
                const { documents, diagnostics } = yield this.loadDocuments(uri, script.runStmts);
                // add diagnostics related to the actual load
                result.loadDiagnostics.push(...diagnostics.map(error => serverUtils_1.addDiagnosticsUri(error, uri)));
                // for each document run validate and yield any results
                for (const document of documents) {
                    const cached = this.documentInfos.get(document.uri);
                    if (!typeGuards_1.empty(cached)) {
                        result.documentInfos.push(cached);
                    }
                    else {
                        const documentInfo = yield this.validateDocument_(document.uri, document.getText(), depth + 1);
                        // if document valid cache and push to result
                        if (!typeGuards_1.empty(documentInfo)) {
                            this.documentInfos.set(document.uri, documentInfo);
                            result.documentInfos.push(documentInfo);
                        }
                    }
                }
            }
            return result;
        });
    }
    /**
     * Get all valid uris from the documents run statements
     * @param uri uri of the calling document
     * @param runStmts run statements in the document
     */
    loadDocuments(uri, runStmts) {
        return __awaiter(this, void 0, void 0, function* () {
            const documents = [];
            const diagnostics = [];
            for (const runStmt of runStmts) {
                // attempt to get a resolvable path from a run statement
                const path = pathResolver_1.runPath(runStmt);
                if (typeof path === 'string') {
                    // attempt to load document
                    const document = yield this.documentService.loadDocumentFromScript(runStmt.toLocation(uri), path);
                    if (!typeGuards_1.empty(document)) {
                        // determine if document or diagnostic
                        if (vscode_languageserver_1.TextDocument.is(document)) {
                            documents.push(document);
                        }
                        else {
                            diagnostics.push(document);
                        }
                    }
                    // was dynamically loaded path can't load
                }
                else {
                    diagnostics.push(path);
                }
            }
            // generate uris then remove empty or preloaded documents
            return {
                documents,
                diagnostics,
            };
        });
    }
    /**
     * Get the symbol table corresponding to the active standard library. For now this
     * will just be the default library but this may allow us to eventually support multiple
     * version of kOS
     */
    activeStandardLibrary() {
        return this.standardLibrary;
    }
    /**
     * Get the symbol table corresponding active set of celestial bodies for the user.
     * This allows for bodies other than that in stock ksp to be incorporated
     */
    activeBodyLibrary() {
        return this.bodyLibrary;
    }
}
exports.AnalysisService = AnalysisService;
//# sourceMappingURL=analysisService.js.map