import { IConnection, TextDocument, Diagnostic, Location } from 'vscode-languageserver';
import { URI } from 'vscode-uri';
import { DocumentLoader, Document } from '../utilities/documentLoader';
declare type DocumentChangeHandler = (document: Document) => void;
declare type DocumentClosedHandler = (uri: string) => void;
declare type DocumentConnection = Pick<IConnection, 'onDidChangeTextDocument' | 'onDidCloseTextDocument' | 'onDidOpenTextDocument'>;
/**
 * Service responsible for managing documents being loaded by the client and the server
 */
export declare class DocumentService {
    /**
     * Currently cached documents loaded by the client
     */
    private clientDocs;
    /**
     * Current cached document only loaded by the server
     */
    private serverDocs;
    /**
     * client connection with events for open, close, and change events
     */
    private conn;
    /**
     * A logger to log information to
     */
    private logger;
    /**
     * A tracer to location exception
     */
    private tracer;
    /**
     * A function to load a uri asynchronously
     */
    private documentLoader;
    /**
     * The path resolver to identifying file paths from kos run paths
     */
    private pathResolver;
    /**
     * Create a new instance of the document service.
     * @param conn document connection holding the required callbacks from iconnection
     * @param uriLoader service to load from a provided uri
     * @param logger logger to log messages to client
     * @param tracer tracer to location exception
     * @param volume0Uri the uri to volume 0 on the drive
     */
    constructor(conn: DocumentConnection, uriLoader: DocumentLoader, logger: ILogger, tracer: ITracer, volume0Uri?: string);
    /**
     * Is the document service read
     */
    ready(): boolean;
    /**
     * Set the volume 0 uri
     * @param uri uri of volume 0
     */
    setVolume0Uri(uri: URI): Promise<void>;
    /**
     * Get a document from this document service
     * @param uri uri to lookup document
     */
    getDocument(uri: string): Maybe<TextDocument>;
    /**
     * Get all document held in this service
     */
    getAllDocuments(): TextDocument[];
    /**
     * Attempt load a document from a kOS script
     * @param caller the caller location for the document
     * @param kosPath the path in the run statement
     */
    loadDocumentFromScript(caller: Location, kosPath: string): Promise<Maybe<Diagnostic | TextDocument>>;
    /**
     * Attempt to load a document from a provided uri
     * @param uri the requested uri
     */
    loadDocument(uri: string): Promise<Maybe<TextDocument>>;
    /**
     * Called when a document is changed. In this is also called when the editor opens a document
     * in order to synchronize
     * @param handler document change handler
     */
    onChange(handler: DocumentChangeHandler): void;
    /**
     * Called when the editor closes a document.
     * @param handler document close handler
     */
    onClose(handler: DocumentClosedHandler): void;
    /**
     * Cache all documents in the workspace.
     */
    private cacheDocuments;
    /**
     * Generate a loading diagnostic if file cannot be loaded
     * @param range range of the run statement
     * @param path kos path of the file
     */
    private loadError;
    /**
     * Retrieve a resource from the provided uri
     * @param uri uri to load resources from
     */
    private retrieveResource;
    /**
     * Handle calls to open text document.
     * @param handler handler to call when did open text document events fires
     */
    private onOpenHandler;
    /**
     * Handle calls to on change text document
     * @param handler handler to call when did change text document event fires
     */
    private onChangeHandler;
    /**
     * On document close handler. Move document from editor docs to server docs.
     * This call notifies any provided handler
     * @param handler document closed handler
     */
    private onCloseHandler;
}
export {};
