"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const typeGuards_1 = require("../utilities/typeGuards");
const pathResolver_1 = require("../utilities/pathResolver");
const vscode_uri_1 = require("vscode-uri");
const diagnosticsUtils_1 = require("../utilities/diagnosticsUtils");
const logger_1 = require("../utilities/logger");
/**
 * Service responsible for managing documents being loaded by the client and the server
 */
class DocumentService {
    /**
     * Create a new instance of the document service.
     * @param conn document connection holding the required callbacks from iconnection
     * @param uriLoader service to load from a provided uri
     * @param logger logger to log messages to client
     * @param tracer tracer to location exception
     * @param volume0Uri the uri to volume 0 on the drive
     */
    constructor(conn, uriLoader, logger, tracer, volume0Uri) {
        this.clientDocs = new Map();
        this.serverDocs = new Map();
        this.pathResolver = new pathResolver_1.PathResolver(volume0Uri);
        this.conn = conn;
        this.documentLoader = uriLoader;
        this.logger = logger;
        this.tracer = tracer;
    }
    /**
     * Is the document service read
     */
    ready() {
        return this.pathResolver.ready();
    }
    /**
     * Set the volume 0 uri
     * @param uri uri of volume 0
     */
    setVolume0Uri(uri) {
        return __awaiter(this, void 0, void 0, function* () {
            this.pathResolver.volume0Uri = uri;
            this.cacheDocuments();
        });
    }
    /**
     * Get a document from this document service
     * @param uri uri to lookup document
     */
    getDocument(uri) {
        const normalized = pathResolver_1.normalizeExtensions(uri);
        if (typeGuards_1.empty(normalized)) {
            return undefined;
        }
        return this.clientDocs.get(normalized) || this.serverDocs.get(normalized);
    }
    /**
     * Get all document held in this service
     */
    getAllDocuments() {
        return [...this.clientDocs.values(), ...this.serverDocs.values()];
    }
    /**
     * Attempt load a document from a kOS script
     * @param caller the caller location for the document
     * @param kosPath the path in the run statement
     */
    loadDocumentFromScript(caller, kosPath) {
        return __awaiter(this, void 0, void 0, function* () {
            // resolver must first be ready
            if (!this.ready()) {
                return undefined;
            }
            // resolve kos path to uri
            const uri = this.pathResolver.resolveUri(caller, kosPath);
            if (typeGuards_1.empty(uri)) {
                return this.loadError(caller.range, kosPath);
            }
            // attempt to load a resource from whatever uri is provided
            const normalized = pathResolver_1.normalizeExtensions(uri);
            if (typeGuards_1.empty(normalized)) {
                return this.loadError(caller.range, kosPath);
            }
            // check for cached versions first
            const cached = this.serverDocs.get(normalized);
            if (!typeGuards_1.empty(cached)) {
                return cached;
            }
            try {
                // attempt to load a resource from whatever uri is provided
                const uri = vscode_uri_1.URI.parse(normalized);
                const retrieved = yield this.retrieveResource(uri);
                const textDocument = vscode_languageserver_1.TextDocument.create(normalized, 'kos', 0, retrieved);
                // if found set in cache and return document
                this.serverDocs.set(normalized, textDocument);
                return textDocument;
            }
            catch (err) {
                // create a diagnostic if we can't load the file
                return this.loadError(caller.range, kosPath);
            }
        });
    }
    /**
     * Attempt to load a document from a provided uri
     * @param uri the requested uri
     */
    loadDocument(uri) {
        return __awaiter(this, void 0, void 0, function* () {
            // resolver must first be ready
            if (!this.ready()) {
                return undefined;
            }
            // attempt to load a resource from whatever uri is provided
            const normalized = pathResolver_1.normalizeExtensions(uri);
            if (typeGuards_1.empty(normalized)) {
                return undefined;
            }
            // check for cached versions first
            const cached = this.serverDocs.get(normalized);
            if (!typeGuards_1.empty(cached)) {
                return cached;
            }
            try {
                // attempt to load a resource from whatever uri is provided
                const uri = vscode_uri_1.URI.parse(normalized);
                const retrieved = yield this.retrieveResource(uri);
                const textDocument = vscode_languageserver_1.TextDocument.create(normalized, 'kos', 0, retrieved);
                // if found set in cache and return document
                this.serverDocs.set(normalized, textDocument);
                return textDocument;
            }
            catch (err) {
                logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
                return undefined;
            }
        });
    }
    /**
     * Called when a document is changed. In this is also called when the editor opens a document
     * in order to synchronize
     * @param handler document change handler
     */
    onChange(handler) {
        this.onChangeHandler(handler);
        this.onOpenHandler(handler);
    }
    /**
     * Called when the editor closes a document.
     * @param handler document close handler
     */
    onClose(handler) {
        this.onCloseHandler(handler);
    }
    /**
     * Cache all documents in the workspace.
     */
    cacheDocuments() {
        var e_1, _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (typeGuards_1.empty(this.pathResolver.volume0Uri)) {
                return;
            }
            const volume0Path = this.pathResolver.volume0Uri.fsPath;
            try {
                try {
                    for (var _b = __asyncValues(this.documentLoader.loadDirectory(volume0Path)), _c; _c = yield _b.next(), !_c.done;) {
                        const { uri, text } = _c.value;
                        const textDocument = vscode_languageserver_1.TextDocument.create(uri, 'kos', 0, text);
                        if (!this.clientDocs.has(uri)) {
                            this.serverDocs.set(uri, textDocument);
                        }
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (_c && !_c.done && (_a = _b.return)) yield _a.call(_b);
                    }
                    finally { if (e_1) throw e_1.error; }
                }
            }
            catch (err) {
                logger_1.logException(this.logger, logger_1.mockTracer, err, 4 /* error */);
            }
        });
    }
    /**
     * Generate a loading diagnostic if file cannot be loaded
     * @param range range of the run statement
     * @param path kos path of the file
     */
    loadError(range, path) {
        return diagnosticsUtils_1.createDiagnostic(range, `Unable to load script at ${path}`, vscode_languageserver_1.DiagnosticSeverity.Information);
    }
    /**
     * Retrieve a resource from the provided uri
     * @param uri uri to load resources from
     */
    retrieveResource(uri) {
        const path = uri.fsPath;
        return this.documentLoader.load(path);
    }
    /**
     * Handle calls to open text document.
     * @param handler handler to call when did open text document events fires
     */
    onOpenHandler(handler) {
        this.conn.onDidOpenTextDocument(params => {
            const document = params.textDocument;
            this.clientDocs.delete(document.uri);
            this.clientDocs.set(document.uri, vscode_languageserver_1.TextDocument.create(document.uri, document.languageId, document.version, document.text));
            handler(document);
        });
    }
    /**
     * Handle calls to on change text document
     * @param handler handler to call when did change text document event fires
     */
    onChangeHandler(handler) {
        this.conn.onDidChangeTextDocument(params => {
            // lookup existing document
            const document = this.clientDocs.get(params.textDocument.uri);
            if (typeGuards_1.empty(document)) {
                return;
            }
            // find all edits that have defined range
            const edits = [];
            for (const change of params.contentChanges) {
                // TODO can't find instance where range is undefined
                if (typeGuards_1.empty(change.range)) {
                    this.logger.error('Document context change had undefined range for this change');
                    this.logger.error(JSON.stringify(change));
                    continue;
                }
                edits.push({ range: change.range, newText: change.text });
            }
            // apply edits
            const text = vscode_languageserver_1.TextDocument.applyEdits(document, edits);
            // create new document
            const updatedDoc = vscode_languageserver_1.TextDocument.create(document.uri, document.languageId, document.version, text);
            // update editor docs
            this.clientDocs.set(document.uri, updatedDoc);
            this.serverDocs.delete(document.uri);
            // call handler
            handler({ text, uri: document.uri });
        });
    }
    /**
     * On document close handler. Move document from editor docs to server docs.
     * This call notifies any provided handler
     * @param handler document closed handler
     */
    onCloseHandler(handler) {
        this.conn.onDidCloseTextDocument(params => {
            const { uri } = params.textDocument;
            // get the current document and remove
            const document = this.clientDocs.get(uri);
            this.clientDocs.delete(uri);
            // if document not empty add to server docs
            if (!typeGuards_1.empty(document)) {
                this.serverDocs.set(uri, document);
            }
            // handle document closing
            handler(uri);
        });
    }
}
exports.DocumentService = DocumentService;
//# sourceMappingURL=documentService.js.map