"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
const vscode_languageserver_1 = require("vscode-languageserver");
const tokentypes_1 = require("../entities/tokentypes");
const treeTraverse_1 = require("../parser/treeTraverse");
/**
 * A service for identifying foldable regions inside a kerboscript
 */
class FoldableService extends treeTraverse_1.TreeTraverse {
    /**
     * Construct a foldable service
     */
    constructor() {
        super();
        this.result = [];
    }
    /**
     * Find foldable regions within a document info
     * @param script the script ast
     * @param regions the region token throughout the script
     */
    findRegions(script, regions) {
        this.result = [];
        for (const stmt of script.stmts) {
            this.stmtAction(stmt);
        }
        this.foldableRegions(regions);
        return [...this.result];
    }
    /**
     * Visit block statements to identify when folding regions are present
     * @param stmt block statement
     */
    visitBlock(stmt) {
        this.result.push({
            startCharacter: stmt.start.character,
            startLine: stmt.start.line,
            endCharacter: stmt.end.character,
            endLine: stmt.end.line,
            kind: vscode_languageserver_1.FoldingRangeKind.Region,
        });
        for (const childStmt of stmt.stmts) {
            this.stmtAction(childStmt);
        }
    }
    /**
     * What are the foldable regions within this document using `\\ #region` and `\\ #endregion`
     * @param regions regions tokens
     */
    foldableRegions(regions) {
        const regionStack = [];
        for (const region of regions) {
            if (region.type === tokentypes_1.TokenType.region) {
                regionStack.push(region);
            }
            else {
                const beginRegion = regionStack.pop();
                if (!typeGuards_1.empty(beginRegion)) {
                    this.result.push({
                        startCharacter: beginRegion.start.character,
                        startLine: beginRegion.start.line,
                        endCharacter: region.end.character,
                        endLine: region.end.line,
                        kind: vscode_languageserver_1.FoldingRangeKind.Region,
                    });
                }
            }
        }
    }
}
exports.FoldableService = FoldableService;
//# sourceMappingURL=foldableService.js.map