"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const documentService_1 = require("../services/documentService");
const vscode_languageserver_1 = require("vscode-languageserver");
const logger_1 = require("../utilities/logger");
const vscode_uri_1 = require("vscode-uri");
const typeGuards_1 = require("../utilities/typeGuards");
const arrayUtils_1 = require("../utilities/arrayUtils");
const path_1 = require("path");
const scanner_1 = require("../scanner/scanner");
const parser_1 = require("../parser/parser");
const foldableService_1 = require("../services/foldableService");
const mockServices_1 = require("./utilities/mockServices");
const positionUtils_1 = require("../utilities/positionUtils");
const analysisService_1 = require("../services/analysisService");
const initialize_1 = require("../typeChecker/initialize");
initialize_1.typeInitializer();
describe('documentService', () => {
    test('ready', () => __awaiter(this, void 0, void 0, function* () {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const baseUri = vscode_uri_1.URI.file('/example/folder');
        const callingUri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer);
        const documentLoaded = yield docService.loadDocumentFromScript({
            uri: callingUri,
            range: {
                start: {
                    line: 0,
                    character: 0,
                },
                end: {
                    line: 1,
                    character: 0,
                },
            },
        }, callingUri);
        expect(docService.ready()).toBe(false);
        expect(documentLoaded).toBeUndefined();
        docService.setVolume0Uri(baseUri);
        expect(docService.ready()).toBe(true);
    }));
    test('load from client', () => {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer);
        const serverDocs = docService['serverDocs'];
        const clientDocs = docService['clientDocs'];
        let i = 0;
        const uris = [
            vscode_uri_1.URI.file('/example/doc1.ks'),
            vscode_uri_1.URI.file('/example/doc2.ks'),
            vscode_uri_1.URI.file('/example/doc3.ks'),
            vscode_uri_1.URI.file('/example/doc4.ks'),
        ];
        const docs = ['example 1', 'example 2', 'example 3', 'example 4'];
        docService.onChange(document => {
            expect(document.uri).toBe(uris[i].toString());
            expect(document.text).toBe(docs[i]);
        });
        for (i = 0; i < uris.length; i += 1) {
            mockConnection.callOpen({
                textDocument: vscode_languageserver_1.TextDocumentItem.create(uris[i].toString(), 'kos', 1, docs[i]),
            });
            expect(serverDocs.size).toBe(0);
            expect(clientDocs.size).toBe(i + 1);
            for (let j = 0; j <= i; j += 1) {
                const doc = docService.getDocument(uris[j].toString());
                expect(doc).toBeDefined();
                if (!typeGuards_1.empty(doc)) {
                    expect(doc.getText()).toBe(docs[j]);
                }
                expect(clientDocs.has(uris[j].toString())).toBe(true);
            }
        }
        const documents = docService.getAllDocuments();
        for (const document of documents) {
            let found = false;
            for (const [uri, doc] of arrayUtils_1.zip(uris, docs)) {
                if (uri.toString() === document.uri && doc === document.getText()) {
                    found = true;
                }
            }
            expect(found).toBe(true);
        }
    });
    test('load from server using kOS run', () => __awaiter(this, void 0, void 0, function* () {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        const callingUri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer, baseUri);
        const serverDocs = docService['serverDocs'];
        const clientDocs = docService['clientDocs'];
        const uris = [
            vscode_uri_1.URI.file('/example/folder/doc1.ks'),
            vscode_uri_1.URI.file('/example/folder/doc2.ks'),
            vscode_uri_1.URI.file('/example/folder/doc3.ks'),
            vscode_uri_1.URI.file('/example/folder/doc4.ks'),
        ];
        const docs = ['example 1', 'example 2', 'example 3', 'example 4'];
        let i = 0;
        for (const [uri, doc] of arrayUtils_1.zip(uris, docs)) {
            const uriString = uri.toString();
            files.set(uri.fsPath, doc);
            const loadedDoc = yield docService.loadDocumentFromScript(vscode_languageserver_1.Location.create(callingUri, vscode_languageserver_1.Range.create({ line: 0, character: 0 }, { line: 0, character: 10 })), `0:/${path_1.basename(uriString)}`);
            expect(loadedDoc).toBeDefined();
            if (!typeGuards_1.empty(loadedDoc)) {
                expect(vscode_languageserver_1.TextDocument.is(loadedDoc)).toBe(true);
                expect(loadedDoc.getText()).toBe(doc);
            }
            expect(clientDocs.size).toBe(0);
            expect(serverDocs.size).toBe(i + 1);
            expect(serverDocs.has(uri.toString())).toBe(true);
            for (let j = 0; j <= i; j += 1) {
                const doc = docService.getDocument(uris[j].toString());
                expect(doc).toBeDefined();
                if (!typeGuards_1.empty(doc)) {
                    expect(doc.getText()).toBe(docs[j]);
                }
                expect(serverDocs.has(uris[j].toString())).toBe(true);
            }
            i = i + 1;
        }
        const nonExistentUri = [
            vscode_uri_1.URI.file('/example/folder/none1.ks'),
            vscode_uri_1.URI.file('/example/folder/none2.ks'),
        ];
        for (const uri of nonExistentUri) {
            const uriString = uri.toString();
            const loadedDoc = yield docService.loadDocumentFromScript(vscode_languageserver_1.Location.create(callingUri, vscode_languageserver_1.Range.create({ line: 0, character: 0 }, { line: 0, character: 10 })), `0:/${path_1.basename(uriString)}`);
            expect(loadedDoc).toBeDefined();
            expect(vscode_languageserver_1.Diagnostic.is(loadedDoc)).toBe(true);
            if (vscode_languageserver_1.Diagnostic.is(loadedDoc)) {
                expect(loadedDoc.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Information);
                expect(positionUtils_1.rangeEqual(loadedDoc.range, vscode_languageserver_1.Range.create({ line: 0, character: 0 }, { line: 0, character: 10 }))).toBe(true);
            }
        }
    }));
    test('load from server using uri', () => __awaiter(this, void 0, void 0, function* () {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer, baseUri);
        const serverDocs = docService['serverDocs'];
        const clientDocs = docService['clientDocs'];
        const uris = [
            vscode_uri_1.URI.file('/example/folder/doc1.ks'),
            vscode_uri_1.URI.file('/example/folder/doc2.ks'),
            vscode_uri_1.URI.file('/example/folder/doc3.ks'),
            vscode_uri_1.URI.file('/example/folder/doc4.ks'),
        ];
        const docs = ['example 1', 'example 2', 'example 3', 'example 4'];
        let i = 0;
        for (const [uri, doc] of arrayUtils_1.zip(uris, docs)) {
            const uriString = uri.toString();
            files.set(uri.fsPath, doc);
            const loadedDoc = yield docService.loadDocument(uriString);
            expect(loadedDoc).toBeDefined();
            if (!typeGuards_1.empty(loadedDoc)) {
                expect(vscode_languageserver_1.TextDocument.is(loadedDoc)).toBe(true);
                expect(loadedDoc.getText()).toBe(doc);
            }
            expect(clientDocs.size).toBe(0);
            expect(serverDocs.size).toBe(i + 1);
            expect(serverDocs.has(uri.toString())).toBe(true);
            for (let j = 0; j <= i; j += 1) {
                const doc = docService.getDocument(uris[j].toString());
                expect(doc).toBeDefined();
                if (!typeGuards_1.empty(doc)) {
                    expect(doc.getText()).toBe(docs[j]);
                }
                expect(serverDocs.has(uris[j].toString())).toBe(true);
            }
            i = i + 1;
        }
        const nonExistentUri = [
            vscode_uri_1.URI.file('/example/folder/none1.ks'),
            vscode_uri_1.URI.file('/example/folder/none2.ks'),
        ];
        for (const uri of nonExistentUri) {
            const uriString = uri.toString();
            const loadedDoc = yield docService.loadDocument(uriString);
            expect(loadedDoc).toBeUndefined();
        }
    }));
    test('change update', () => __awaiter(this, void 0, void 0, function* () {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        // const callingUri = URI.file('/example/folder/example.ks').toString();
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer, baseUri);
        const serverDocs = docService['serverDocs'];
        const clientDocs = docService['clientDocs'];
        const uri = vscode_uri_1.URI.file('/example/folder/doc1.ks');
        const content = 'example';
        const afterEdit = 'example edited';
        let first = true;
        docService.onChange(document => {
            if (first) {
                expect(document.uri).toBe(uri.toString());
                expect(document.text).toBe(content);
                first = false;
            }
            else {
                expect(document.uri).toBe(uri.toString());
                expect(document.text).toBe(afterEdit);
            }
        });
        docService.onClose(closeUri => {
            expect(closeUri).toBe(uri.toString());
        });
        mockConnection.callOpen({
            textDocument: vscode_languageserver_1.TextDocumentItem.create(uri.toString(), 'kos', 1, content),
        });
        expect(serverDocs.size).toBe(0);
        expect(clientDocs.size).toBe(1);
        let clientDoc = docService.getDocument(uri.toString());
        expect(clientDoc).toBeDefined();
        if (!typeGuards_1.empty(clientDoc)) {
            expect(clientDoc.getText()).toBe(content);
        }
        mockConnection.callChange({
            textDocument: vscode_languageserver_1.VersionedTextDocumentIdentifier.create(uri.toString(), 1),
            contentChanges: [
                {
                    range: vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(0, 7), vscode_languageserver_1.Position.create(0, 7)),
                    rangeLength: 0,
                    text: ' edited',
                },
            ],
        });
        expect(serverDocs.size).toBe(0);
        expect(clientDocs.size).toBe(1);
        clientDoc = docService.getDocument(uri.toString());
        expect(clientDoc).toBeDefined();
        if (!typeGuards_1.empty(clientDoc)) {
            expect(clientDoc.getText()).toBe(afterEdit);
        }
        mockConnection.callClose({
            textDocument: vscode_languageserver_1.TextDocumentIdentifier.create(uri.toString()),
        });
        expect(serverDocs.size).toBe(1);
        expect(clientDocs.size).toBe(0);
        clientDoc = docService.getDocument(uri.toString());
        expect(clientDoc).toBeDefined();
        if (!typeGuards_1.empty(clientDoc)) {
            expect(clientDoc.getText()).toBe(afterEdit);
        }
    }));
    test('load extension', () => __awaiter(this, void 0, void 0, function* () {
        const mockConnection = mockServices_1.createMockDocConnection();
        const files = new Map();
        const mockUriResponse = mockServices_1.createMockUriResponse(files);
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        const callingUri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const docService = new documentService_1.DocumentService(mockConnection, mockUriResponse, logger_1.mockLogger, logger_1.mockTracer, baseUri);
        const serverDocs = docService['serverDocs'];
        const clientDocs = docService['clientDocs'];
        const actualUri = [
            vscode_uri_1.URI.file('/example/folder/doc1.ks'),
            vscode_uri_1.URI.file('/example/folder/doc2.ks'),
            vscode_uri_1.URI.file('/example/folder/doc3.ks'),
        ];
        const requestedUri = [
            vscode_uri_1.URI.file('/example/folder/doc1.ks'),
            vscode_uri_1.URI.file('/example/folder/doc2.ksm'),
            vscode_uri_1.URI.file('/example/folder/doc3'),
            vscode_uri_1.URI.file('/example/folder/doc4.txt'),
        ];
        const doc = 'example';
        let i = 0;
        for (const [aUri, rUri] of arrayUtils_1.zip(actualUri, requestedUri)) {
            const uriString = rUri.toString();
            files.set(aUri.fsPath, doc);
            const loadedDoc = yield docService.loadDocumentFromScript(vscode_languageserver_1.Location.create(callingUri, vscode_languageserver_1.Range.create({ line: 0, character: 0 }, { line: 0, character: 10 })), `0:/${path_1.basename(uriString)}`);
            expect(loadedDoc).toBeDefined();
            if (!typeGuards_1.empty(loadedDoc)) {
                expect(vscode_languageserver_1.TextDocument.is(loadedDoc)).toBe(true);
                expect(loadedDoc.getText()).toBe(doc);
            }
            expect(clientDocs.size).toBe(0);
            expect(serverDocs.size).toBe(i + 1);
            expect(serverDocs.has(aUri.toString())).toBe(true);
            i += 1;
        }
    }));
});
describe('analysisService', () => {
    test('validate single document', () => __awaiter(this, void 0, void 0, function* () {
        const uri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const documents = new Map([
            [uri, vscode_languageserver_1.TextDocument.create(uri, 'kos', 1.0, 'print(10).')],
        ]);
        const docService = mockServices_1.createMockDocumentService(documents, vscode_uri_1.URI.file('/').toString());
        const analysisService = new analysisService_1.AnalysisService(2 /* camelcase */, logger_1.mockLogger, logger_1.mockTracer, docService);
        const diagnostics = yield analysisService.validateDocument(uri, documents.get(uri).getText());
        const documentInfo = yield analysisService.getInfo(uri);
        expect(diagnostics.length).toBe(0);
        expect(documentInfo).toBeDefined();
        if (!typeGuards_1.empty(documentInfo)) {
            expect(documentInfo.symbolTable.dependencyTables.size).toBe(2);
            expect(documentInfo.diagnostics).toStrictEqual(diagnostics);
            expect(documentInfo.script.stmts.length).toBe(1);
            expect(documentInfo.symbolTable.rootScope.environment.symbols.length).toBe(0);
        }
    }));
    test('validate single document getinfo first', () => __awaiter(this, void 0, void 0, function* () {
        const uri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const documents = new Map([
            [uri, vscode_languageserver_1.TextDocument.create(uri, 'kos', 1.0, 'print(10).')],
        ]);
        const docService = mockServices_1.createMockDocumentService(documents, vscode_uri_1.URI.file('/').toString());
        const analysisService = new analysisService_1.AnalysisService(2 /* camelcase */, logger_1.mockLogger, logger_1.mockTracer, docService);
        const documentInfo = yield analysisService.getInfo(uri);
        expect(documentInfo).toBeDefined();
        if (!typeGuards_1.empty(documentInfo)) {
            expect(documentInfo.symbolTable.dependencyTables.size).toBe(2);
            expect(documentInfo.script.stmts.length).toBe(1);
            expect(documentInfo.symbolTable.rootScope.environment.symbols.length).toBe(0);
        }
        const diagnostics = yield analysisService.validateDocument(uri, documents.get(uri).getText());
        expect(diagnostics.length).toBe(0);
        if (!typeGuards_1.empty(documentInfo)) {
            expect(documentInfo.symbolTable.dependencyTables.size).toBe(0);
            expect(documentInfo.diagnostics).toStrictEqual(diagnostics);
            expect(documentInfo.script.stmts.length).toBe(1);
            expect(documentInfo.symbolTable.rootScope.environment.symbols.length).toBe(0);
        }
    }));
    test('set case', () => __awaiter(this, void 0, void 0, function* () {
        const uri = vscode_uri_1.URI.file('/example/folder/example.ks').toString();
        const documents = new Map([
            [uri, vscode_languageserver_1.TextDocument.create(uri, 'kos', 1.0, 'print(10).')],
        ]);
        const docService = mockServices_1.createMockDocumentService(documents, vscode_uri_1.URI.file('/').toString());
        const analysisService = new analysisService_1.AnalysisService(0 /* lowercase */, logger_1.mockLogger, logger_1.mockTracer, docService);
        let bodyLib = analysisService['bodyLibrary'];
        let stdLib = analysisService['bodyLibrary'];
        for (const bodySymbol of bodyLib.allSymbols()) {
            expect(bodySymbol.name.lexeme).toBe(bodySymbol.name.lexeme.toLowerCase());
        }
        for (const stdSymbol of stdLib.allSymbols()) {
            expect(stdSymbol.name.lexeme).toBe(stdSymbol.name.lexeme.toLowerCase());
        }
        analysisService.setCase(1 /* uppercase */);
        bodyLib = analysisService['bodyLibrary'];
        stdLib = analysisService['bodyLibrary'];
        for (const bodySymbol of bodyLib.allSymbols()) {
            expect(bodySymbol.name.lexeme).toBe(bodySymbol.name.lexeme.toUpperCase());
        }
        for (const stdSymbol of stdLib.allSymbols()) {
            expect(stdSymbol.name.lexeme).toBe(stdSymbol.name.lexeme.toUpperCase());
        }
    }));
    test('validate multiple documents', () => __awaiter(this, void 0, void 0, function* () {
        const uri1 = vscode_uri_1.URI.file('/example/folder/example1.ks').toString();
        const uri2 = vscode_uri_1.URI.file('/example/folder/example2.ks').toString();
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        const documents = new Map([
            [
                uri1,
                vscode_languageserver_1.TextDocument.create(uri1, 'kos', 1.0, 'runOncePath("example2.ks"). hi().'),
            ],
            [
                uri2,
                vscode_languageserver_1.TextDocument.create(uri2, 'kos', 1.0, 'function hi { print("hi"). }'),
            ],
        ]);
        const docService = mockServices_1.createMockDocumentService(documents, baseUri);
        const analysisService = new analysisService_1.AnalysisService(2 /* camelcase */, logger_1.mockLogger, logger_1.mockTracer, docService);
        const diagnostics = yield analysisService.validateDocument(uri1, documents.get(uri1).getText());
        const documentInfo1 = yield analysisService.getInfo(uri1);
        const documentInfo2 = yield analysisService.getInfo(uri2);
        expect(diagnostics.length).toBe(0);
        expect(documentInfo1).toBeDefined();
        expect(documentInfo2).toBeDefined();
        if (!typeGuards_1.empty(documentInfo1) && !typeGuards_1.empty(documentInfo2)) {
            expect(documentInfo1.symbolTable.dependencyTables.size).toBe(3);
            expect(documentInfo2.symbolTable.dependencyTables.size).toBe(2);
            expect(documentInfo1.symbolTable.dependencyTables).toContain(documentInfo2.symbolTable);
            expect(documentInfo1.diagnostics).toStrictEqual(diagnostics);
            expect(documentInfo1.script.stmts.length).toBe(2);
            expect(documentInfo2.script.stmts.length).toBe(1);
            expect(documentInfo1.symbolTable.rootScope.environment.symbols().length).toBe(0);
            expect(documentInfo2.symbolTable.rootScope.environment.symbols().length).toBe(1);
        }
    }));
    test('validate multiple with updates documents', () => __awaiter(this, void 0, void 0, function* () {
        const uri1 = vscode_uri_1.URI.file('/example/folder/example1.ks').toString();
        const uri2 = vscode_uri_1.URI.file('/example/folder/example2.ks').toString();
        const baseUri = vscode_uri_1.URI.file('/example/folder').toString();
        const documents = new Map([
            [
                uri1,
                vscode_languageserver_1.TextDocument.create(uri1, 'kos', 1.0, 'runOncePath("example2.ks"). hi().'),
            ],
            [
                uri2,
                vscode_languageserver_1.TextDocument.create(uri2, 'kos', 1.0, 'function hi { print("hi"). }'),
            ],
        ]);
        const docService = mockServices_1.createMockDocumentService(documents, baseUri);
        const analysisService = new analysisService_1.AnalysisService(2 /* camelcase */, logger_1.mockLogger, logger_1.mockTracer, docService);
        // initial load of example1.ks
        const diagnostics11 = yield analysisService.validateDocument(uri1, documents.get(uri1).getText());
        const documentInfo11 = yield analysisService.getInfo(uri1);
        // load from client example2.ks
        const diagnostics21 = yield analysisService.validateDocument(uri2, documents.get(uri2).getText());
        const documentInfo21 = yield analysisService.getInfo(uri2);
        // update load of example1.ks
        const diagnostics12 = yield analysisService.validateDocument(uri1, documents.get(uri1).getText());
        const documentInfo12 = yield analysisService.getInfo(uri1);
        // update load of example2.ks
        const diagnostics22 = yield analysisService.validateDocument(uri2, documents.get(uri2).getText());
        const documentInfo22 = yield analysisService.getInfo(uri2);
        expect(diagnostics11.length).toBe(0);
        expect(diagnostics12.length).toBe(0);
        expect(diagnostics21.length).toBe(0);
        expect(diagnostics22.length).toBe(0);
        expect(documentInfo11).toBeDefined();
        expect(documentInfo12).toBeDefined();
        expect(documentInfo21).toBeDefined();
        expect(documentInfo22).toBeDefined();
        const documentInfos = analysisService['documentInfos'];
        if (!typeGuards_1.empty(documentInfo11) && !typeGuards_1.empty(documentInfo21)) {
            expect(documentInfo11.symbolTable.dependencyTables.size).toBe(0);
            expect(documentInfo11.symbolTable.dependentTables.size).toBe(0);
            expect(documentInfo21.symbolTable.dependencyTables.size).toBe(0);
            expect(documentInfo21.symbolTable.dependentTables.size).toBe(0);
            expect(documentInfo11.diagnostics).toStrictEqual(diagnostics11);
            expect(documentInfo11.diagnostics).toStrictEqual(diagnostics12);
            expect(documentInfo21.diagnostics).toStrictEqual(diagnostics21);
            expect(documentInfo21.diagnostics).toStrictEqual(diagnostics22);
            expect(documentInfos.get(uri1)).not.toBe(documentInfo11);
            expect(documentInfos.get(uri2)).not.toBe(documentInfo21);
        }
        if (!typeGuards_1.empty(documentInfo12) && !typeGuards_1.empty(documentInfo22)) {
            expect(documentInfo12.symbolTable.dependencyTables.size).toBe(3);
            expect(documentInfo12.symbolTable.dependentTables.size).toBe(0);
            expect(documentInfo22.symbolTable.dependencyTables.size).toBe(2);
            expect(documentInfo22.symbolTable.dependentTables.size).toBe(1);
            expect(documentInfos.get(uri1)).toBe(documentInfo12);
            expect(documentInfos.get(uri2)).toBe(documentInfo22);
        }
    }));
});
// #region string scripts
const regionFold = `
// #region

// #endregion
`;
const blockFold = `
if true {

}

function example {
  print("hi").
}
`;
const bothFold = `
// #region
if true {

}

function example {
  print("hi").
}
// #endregion
`;
// #endregion
const fakeUri = 'file:///fake.ks';
// parse source
const parseSource = (source) => {
    const scanner = new scanner_1.Scanner(source, fakeUri);
    const scan = scanner.scanTokens();
    const parser = new parser_1.Parser(fakeUri, scan.tokens);
    const parse = parser.parse();
    return { scan, parse };
};
const noParseErrors = (result) => {
    expect(result.scan.scanErrors.length).toBe(0);
    expect(result.parse.parseErrors.length).toBe(0);
};
describe('foldableService', () => {
    test('Fold region', () => {
        const result = parseSource(regionFold);
        noParseErrors(result);
        const service = new foldableService_1.FoldableService();
        const foldable = service.findRegions(result.parse.script, result.scan.regions);
        const folds = [
            {
                startCharacter: 0,
                startLine: 1,
                endCharacter: 13,
                endLine: 3,
                kind: 'region',
            },
        ];
        expect(foldable).toContainEqual(folds[0]);
    });
    test('Fold block', () => {
        const result = parseSource(blockFold);
        noParseErrors(result);
        const service = new foldableService_1.FoldableService();
        const foldable = service.findRegions(result.parse.script, result.scan.regions);
        expect(foldable.length).toBe(2);
        const folds = [
            {
                startCharacter: 8,
                startLine: 1,
                endCharacter: 1,
                endLine: 3,
                kind: 'region',
            },
            {
                startCharacter: 17,
                startLine: 5,
                endCharacter: 1,
                endLine: 7,
                kind: 'region',
            },
        ];
        expect(foldable).toContainEqual(folds[0]);
        expect(foldable).toContainEqual(folds[1]);
    });
    test('Fold both', () => {
        const result = parseSource(bothFold);
        noParseErrors(result);
        const service = new foldableService_1.FoldableService();
        const foldable = service.findRegions(result.parse.script, result.scan.regions);
        expect(foldable.length).toBe(3);
        const folds = [
            {
                startCharacter: 8,
                startLine: 2,
                endCharacter: 1,
                endLine: 4,
                kind: 'region',
            },
            {
                startCharacter: 17,
                startLine: 6,
                endCharacter: 1,
                endLine: 8,
                kind: 'region',
            },
            {
                startCharacter: 0,
                startLine: 1,
                endCharacter: 13,
                endLine: 9,
                kind: 'region',
            },
        ];
        expect(foldable).toContainEqual(folds[0]);
        expect(foldable).toContainEqual(folds[1]);
        expect(foldable).toContainEqual(folds[2]);
    });
});
//# sourceMappingURL=service.spec.js.map