"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const scanner_1 = require("../scanner/scanner");
const parser_1 = require("../parser/parser");
const symbolTableBuilder_1 = require("../analysis/symbolTableBuilder");
const standardLibrary_1 = require("../analysis/standardLibrary");
const preResolver_1 = require("../analysis/preResolver");
const resolver_1 = require("../analysis/resolver");
const typeChecker_1 = require("../typeChecker/typeChecker");
const types_1 = require("../analysis/types");
const typeGuards_1 = require("../utilities/typeGuards");
const boolean_1 = require("../typeChecker/ksTypes/primitives/boolean");
const scalar_1 = require("../typeChecker/ksTypes/primitives/scalar");
const string_1 = require("../typeChecker/ksTypes/primitives/string");
const userList_1 = require("../typeChecker/ksTypes/collections/userList");
const structure_1 = require("../typeChecker/ksTypes/primitives/structure");
const vector_1 = require("../typeChecker/ksTypes/collections/vector");
const direction_1 = require("../typeChecker/ksTypes/collections/direction");
const marker_1 = require("../entities/marker");
const arrayUtils_1 = require("../utilities/arrayUtils");
const timespan_1 = require("../typeChecker/ksTypes/timespan");
const initialize_1 = require("../typeChecker/initialize");
const bodyatmosphere_1 = require("../typeChecker/ksTypes/bodyatmosphere");
const list_1 = require("../typeChecker/ksTypes/collections/list");
const part_1 = require("../typeChecker/ksTypes/parts/part");
const path_1 = require("../typeChecker/ksTypes/io/path");
const fakeUri = 'C:\\fake.ks';
initialize_1.typeInitializer();
// parse source
const parseSource = (source) => {
    const scanner = new scanner_1.Scanner(source, fakeUri);
    const scan = scanner.scanTokens();
    const parser = new parser_1.Parser(fakeUri, scan.tokens);
    const parse = parser.parse();
    return { scan, parse };
};
const checkSource = (source, standardLib = false) => {
    const result = parseSource(source);
    const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder(fakeUri);
    if (standardLib) {
        symbolTableBuilder.linkDependency(standardLibrary_1.standardLibraryBuilder(0 /* lowercase */));
    }
    const functionResolver = new preResolver_1.PreResolver(result.parse.script, symbolTableBuilder);
    const resolver = new resolver_1.Resolver(result.parse.script, symbolTableBuilder);
    const preResolverError = functionResolver.resolve();
    const resolverErrors = resolver.resolve();
    const unusedErrors = symbolTableBuilder.findUnused();
    const checker = new typeChecker_1.TypeChecker(result.parse.script);
    const typeCheckError = checker.check();
    return Object.assign({}, result, { resolveDiagnostics: preResolverError.concat(resolverErrors, unusedErrors), typeCheckDiagnostics: typeCheckError, table: symbolTableBuilder.build() });
};
const noResolverErrors = (result) => {
    expect(result.scan.scanErrors.map(e => e.message)).toEqual([]);
    expect(result.parse.parseErrors.map(e => e.message)).toEqual([]);
    expect(result.resolveDiagnostics.map(e => e.message)).toEqual([]);
};
const noErrors = (result) => {
    expect(result.scan.scanErrors.map(e => e.message)).toEqual([]);
    expect(result.parse.parseErrors.map(e => e.message)).toEqual([]);
    expect(result.resolveDiagnostics.map(e => e.message)).toEqual([]);
    expect(result.typeCheckDiagnostics.map(e => e.message)).toEqual([]);
};
const literalSource = `
local b1 is true.
local b2 is false.

local d1 is 10.0.
local d2 is 4e6.

local i is 5.

local fi is file.ks.
local s is "example".

print fi.
print b1.
print i.
print b2.
print d1.
print d2.
print s.
`;
const identiferSource = `
local b1 is true.
local b2 is false.

local d1 is 10.0.
local d2 is 4e6.

local i1 is 5.

local fi1 is file.ks.
local s1 is "example".

local b3 is b1.
local b4 is b2.

local d3 is d1.
local d4 is d2.

local i2 is i1.

local fi2 is fi1.
local s2 is s1.

print fi2.
print i2.
print b3.
print b4.
print d3.
print d4.
print s2.
`;
const collectionSource = `
local a is 0.
local b is "cat".
local c is false.

local l1 is list(1, 2, 3).
local l2 is list(a, b, c).

local x1 is l1[1].
local x2 is l2[1].

print(x1).
print(x2).

for i1 in l1 { print(i1). }
for i2 in l2 { print(i2). }

// need to have userlisttype subtype listtype
local p is path("example").
local segments is p:segments.

local x3 is segments[0].
print(x3).

for i3 in segments { print(i3). }
`;
const suffixSource = `
local atm is body:atm.
local length is list():length.
local parts is ship:parts.
local partFacing is ship:parts[0]:facing.
local distance is body:geopositionlatlng(10, 10):distance.

print(atm).
print(length).
print(parts).
print(partFacing).
print(distance).
`;
const symbolTests = (symbols, name, symbolKind, targetType) => {
    expect(symbols.has(name)).toBe(true);
    const nameWrap = symbols.get(name);
    expect(nameWrap).toBeDefined();
    const nameUnWrap = typeGuards_1.unWrap(nameWrap);
    expect(nameUnWrap.name.lexeme).toBe(name);
    expect(nameUnWrap.tag).toBe(symbolKind);
    expect(nameUnWrap.name.tracker).toBeDefined();
    if (!typeGuards_1.empty(targetType)) {
        const nameTrack = typeGuards_1.unWrap(nameUnWrap.name.tracker);
        expect(nameTrack.declared.type).toBe(targetType);
    }
};
describe('Basic inferring', () => {
    test('Literal inferring', () => {
        const results = checkSource(literalSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'b1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b2', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd1', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'd2', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'i', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 's', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'fi', types_1.KsSymbolKind.variable, string_1.stringType);
    });
    test('Identifier inferring', () => {
        const results = checkSource(identiferSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'b3', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b4', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd3', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'd4', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'i2', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 's2', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'fi2', types_1.KsSymbolKind.variable, string_1.stringType);
    });
    test('collection inferring', () => {
        const results = checkSource(collectionSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'a', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 'b', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'c', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'l1', types_1.KsSymbolKind.variable, userList_1.userListType);
        symbolTests(names, 'l2', types_1.KsSymbolKind.variable, userList_1.userListType);
        symbolTests(names, 'x1', types_1.KsSymbolKind.variable, structure_1.structureType);
        symbolTests(names, 'x2', types_1.KsSymbolKind.variable, structure_1.structureType);
        symbolTests(names, 'i1', types_1.KsSymbolKind.variable, structure_1.structureType);
        symbolTests(names, 'i2', types_1.KsSymbolKind.variable, structure_1.structureType);
        symbolTests(names, 'p', types_1.KsSymbolKind.variable, path_1.pathType);
        symbolTests(names, 'segments', types_1.KsSymbolKind.variable, list_1.listType.apply(string_1.stringType));
        symbolTests(names, 'x3', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'i3', types_1.KsSymbolKind.variable, string_1.stringType);
    });
    test('suffix inferring', () => {
        const results = checkSource(suffixSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'atm', types_1.KsSymbolKind.variable, bodyatmosphere_1.bodyAtmosphereType);
        symbolTests(names, 'length', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 'parts', types_1.KsSymbolKind.variable, list_1.listType.apply(part_1.partType));
        symbolTests(names, 'partFacing', types_1.KsSymbolKind.variable, direction_1.directionType);
        symbolTests(names, 'distance', types_1.KsSymbolKind.variable, scalar_1.scalarType);
    });
});
const unarySource = `
function f { }
local x is 10.
lock l to x.

local d1 is defined f.
local d2 is defined x.
local d3 is defined l.

print(d1).
print(d2).
print(d3).

local b1 is not true.
local b2 is not false.
local b3 is not (10 > 5).

print(b1).
print(b2).
print(b3).

local n1 is -10.
local n2 is -16.3.
local n3 is +18.3.
lock n4 to -v(1, 1, 1).
local n5 is +q(1, 1, 1, 1).

print(n1).
print(n2).
print(n3).
print(n4).
print(n5).
`;
const unaryInvalidSource = `
function f {}.

local b1 is not f.
local n1 is -"test".

print(b1).
print(n1).
`;
const binaryMultiplicationSource = `
local s1 is 10 * 10.

local v1 is v(1, 1, 1) * v(1, 1, 1).
local v2 is v(1, 1, 1) * 10.

local d1 is q(1, 1, 1, 1) * v(1, 1, 1).
local d2 is q(1, 1, 1, 1) * q(1, 1, 1, 1).

local t1 is time * 10.

print(s1).

print(v1).
print(v2).

print(d1).
print(d2).

print(t1).
`;
const binaryAdditionSource = `
local s1 is 10 + 10.
local str1 is "cat" + "dog".

local v1 is v(1, 1, 1) + v(1, 1, 1).

local d1 is q(1, 1, 1, 1) + v(1, 1, 1).
local d2 is q(1, 1, 1, 1) + q(1, 1, 1, 1).

local t1 is time + time.
local t2 is time + 10.

print(s1).
print(str1).

print(v1).

print(d1).
print(d2).

print(t1).
print(t2).
`;
const binaryLogicalSource = `
local l1 is true and true.
local l2 is true or true.
local l3 is ship or true.
local l4 is 10 and 20.
local l5 is "true" or "false".


print(l1).
print(l2).
print(l3).
print(l4).
print(l5).
`;
const factorSource = `
local l1 is 10 ^ 10.
local l2 is 10 ^ 10.0.
local l3 is 10.0 ^ 10.
local l4 is 10.0 ^ 10.0.

local i is 10.
local d is 10.0.

local e1 is i ^ i.
local e2 is i ^ d.
local e3 is d ^ i.
local e4 is d ^ d.

print(l1).
print(l2).
print(l3).
print(l4).

print(e1).
print(e2).
print(e3).
print(e4).
`;
describe('Operators', () => {
    test('unary operators', () => {
        const results = checkSource(unarySource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'f', types_1.KsSymbolKind.function);
        symbolTests(names, 'x', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 'l', types_1.KsSymbolKind.lock, scalar_1.integerType);
        symbolTests(names, 'd1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd2', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd3', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b2', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b3', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'n1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'n2', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'n3', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'n4', types_1.KsSymbolKind.lock, vector_1.vectorType);
        symbolTests(names, 'n5', types_1.KsSymbolKind.variable, direction_1.directionType);
    });
    const unaryLocations = [
        { start: new marker_1.Marker(3, 16), end: new marker_1.Marker(3, 17) },
        { start: new marker_1.Marker(4, 13), end: new marker_1.Marker(4, 19) },
    ];
    test('unary invalid operators', () => {
        const results = checkSource(unaryInvalidSource, true);
        noResolverErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'b1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'n1', types_1.KsSymbolKind.variable, structure_1.structureType);
        const sortedErrors = results.typeCheckDiagnostics.sort((a, b) => a.range.start.line - b.range.start.line);
        for (const [error, location] of arrayUtils_1.zip(sortedErrors, unaryLocations)) {
            expect(error.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Hint);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    test('binary multiplication operators', () => {
        const results = checkSource(binaryMultiplicationSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 's1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'v1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'v2', types_1.KsSymbolKind.variable, vector_1.vectorType);
        symbolTests(names, 'd1', types_1.KsSymbolKind.variable, vector_1.vectorType);
        symbolTests(names, 'd2', types_1.KsSymbolKind.variable, direction_1.directionType);
        symbolTests(names, 't1', types_1.KsSymbolKind.variable, timespan_1.timeSpanType);
    });
    test('binary plus operators', () => {
        const results = checkSource(binaryAdditionSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 's1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'str1', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'v1', types_1.KsSymbolKind.variable, vector_1.vectorType);
        symbolTests(names, 'd1', types_1.KsSymbolKind.variable, vector_1.vectorType);
        symbolTests(names, 'd2', types_1.KsSymbolKind.variable, direction_1.directionType);
        symbolTests(names, 't1', types_1.KsSymbolKind.variable, timespan_1.timeSpanType);
        symbolTests(names, 't2', types_1.KsSymbolKind.variable, timespan_1.timeSpanType);
    });
    test('binary and operators', () => {
        const results = checkSource(binaryLogicalSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'l1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'l2', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'l3', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'l4', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'l5', types_1.KsSymbolKind.variable, boolean_1.booleanType);
    });
    test('factor operator', () => {
        const results = checkSource(factorSource, true);
        noErrors(results);
        const { table } = results;
        const symbols = table.allSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'i', types_1.KsSymbolKind.variable, scalar_1.integerType);
        symbolTests(names, 'd', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'l1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'l2', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'l3', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'l4', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'e1', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'e2', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'e3', types_1.KsSymbolKind.variable, scalar_1.scalarType);
        symbolTests(names, 'e4', types_1.KsSymbolKind.variable, scalar_1.scalarType);
    });
});
//# sourceMappingURL=typeChecking.spec.js.map