"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../../utilities/typeGuards");
const typeCreators_1 = require("../typeCreators");
/**
 * A class that's responsible for binding type arguments to type
 * parameters of parametric call signatures
 */
class Binder {
    /**
     * Construct a new binder
     * @param names names of the types for substituting
     */
    constructor(names) {
        if (new Set(names).size !== names.length) {
            throw new Error(`Type parameter must have unique names. Provided ${names.join(', ')}`);
        }
        this.typeParameters = new Set(names.map(name => typeCreators_1.createTypeParameter(name)));
        this.applicationCache = new Map();
    }
    /**
     * Check if the provided replacement are valid
     * @param typeArguments the requested type substitutions
     */
    areValidArguments(typeArguments) {
        for (const typeParameter of this.typeParameters) {
            if (!typeArguments.has(typeParameter)) {
                throw new Error(`Provided arguments does not have a type for type parameter ${typeParameter.name}`);
            }
        }
    }
    /**
     * Get the cached type arguments to preserve reference equality
     * @param typeReplacement placeholder for type subs
     */
    getCachedArguments(typeReplacement) {
        for (const cachedReplacement of this.applicationCache.keys()) {
            if (cachedReplacement.size !== typeReplacement.size) {
                continue;
            }
            let match = true;
            for (const [key, value] of cachedReplacement.entries()) {
                const found = typeReplacement.get(key);
                if (found !== value) {
                    match = false;
                    break;
                }
            }
            if (match) {
                return cachedReplacement;
            }
        }
        return typeReplacement;
    }
    /**
     * Apply type arguments to this type
     * @param type type to make concrete
     * @param typeArguments type replacement
     */
    applyToType(type, typeArguments) {
        // check if type is actually a parameter
        if (this.typeParameters.has(type)) {
            // if found get arguments for parameter
            const argument = typeArguments.get(type);
            if (typeGuards_1.empty(argument)) {
                throw new Error(`Did not provide a type argument for type parameter for ${type.name}.`);
            }
            return argument;
        }
        // otherwise apply arguments to type
        return type.apply(typeArguments);
    }
    /**
     * Map type argument from this type to one of it's children
     * @param typeArguments type arguments for this type
     * @param parameterMap mapping of type parameters from a source type to a target type
     */
    mapArguments(typeArguments, parameterMap) {
        if (parameterMap.size === 0) {
            return typeArguments;
        }
        const mappedTypeReplacement = new Map();
        for (const [generic, type] of typeArguments) {
            // try to find type parameter from placeholders
            if (!this.typeParameters.has(generic)) {
                throw new Error(`Could not find type parameter for ${generic.name}.`);
            }
            // look up this type's type parameter in the link
            const mappedTypeParam = parameterMap.get(generic);
            if (typeGuards_1.empty(mappedTypeParam)) {
                throw new Error(`Could not find type parameter mapping of ${generic.name}.`);
            }
            // if found map super placeholder to the provided substitution
            mappedTypeReplacement.set(mappedTypeParam, type);
        }
        return mappedTypeReplacement;
    }
}
exports.Binder = Binder;
//# sourceMappingURL=binder.js.map