"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../../utilities/typeGuards");
const callSignature_1 = require("../types/callSignature");
const binder_1 = require("./binder");
/**
 * A class that's responsible for binding type arguments to type
 * parameters of parametric call signatures
 */
class CallBinder extends binder_1.Binder {
    /**
     * Construct a new call binder
     * @param names names of the types for substituting
     */
    constructor(names) {
        super(names);
    }
    /**
     * Apply type arguments to this parametric call signature to generate
     * a call signature
     * @param typeArguments the type arguments for this application
     * @param params the parameters for the generic call
     * @param returns the return of the generic call
     */
    apply(typeArguments, params, returns) {
        // check validity of the substitution
        this.areValidArguments(typeArguments);
        // check if we've already performed this type replacement
        const cachedTypeSubstitutions = this.getCachedArguments(typeArguments);
        const cache = this.applicationCache.get(cachedTypeSubstitutions);
        if (!typeGuards_1.empty(cache)) {
            return cache;
        }
        const newParams = [];
        // convert params
        for (const param of params) {
            const mappedParamSubstitutions = this.mapArguments(typeArguments, param.mapping);
            newParams.push(this.applyToType(param.type, mappedParamSubstitutions));
        }
        const mappedReturnsSubstitutions = this.mapArguments(typeArguments, returns.mapping);
        const newReturns = this.applyToType(returns.type, mappedReturnsSubstitutions);
        // generate the new type
        const newCall = new callSignature_1.CallSignature(newParams, newReturns);
        // cache substitution
        this.applicationCache.set(cachedTypeSubstitutions, newCall);
        return newCall;
    }
}
exports.CallBinder = CallBinder;
//# sourceMappingURL=callBinder.js.map