"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../../utilities/typeGuards");
const operator_1 = require("../types/operator");
const typeCreators_1 = require("../typeCreators");
const type_1 = require("../types/type");
const binder_1 = require("./binder");
/**
 * A class that is responsible for binding type arguments to type parameters in
 * parametric types
 */
class TypeBinder extends binder_1.Binder {
    /**
     * Construct a new type binder
     * @param names names of the type for substituting
     */
    constructor(names) {
        super(names);
    }
    /**
     * Substitute placeholder types in a generic for real one using the
     * provided type arguments
     * @param type the generic type to be made concrete
     * @param typeArguments the type replacement to preform
     * @param callSignature the call signature and it's type mapping
     * @param suffixes the collection of suffixes and their type mapping
     * @param superMap the super and it's type mapping
     */
    apply(type, typeArguments, suffixes, callSignature, superMap, indexerMap) {
        // check validity of the substitution
        this.areValidArguments(typeArguments);
        // check if we've already performed this type replacement
        const cachedTypeReplacement = this.getCachedArguments(typeArguments);
        const cache = this.applicationCache.get(cachedTypeReplacement);
        if (!typeGuards_1.empty(cache)) {
            return cache;
        }
        // update call signature
        let newCallSignature = undefined;
        if (!typeGuards_1.empty(callSignature)) {
            const mappedCallReplacement = this.mapArguments(typeArguments, callSignature.mapping);
            newCallSignature = callSignature.type.apply(mappedCallReplacement);
        }
        // generate the new type
        const newType = new type_1.Type(type.name, type.access, cachedTypeReplacement, type.kind, newCallSignature, type);
        // cache earlier in case of cycles list<T>:copy -> list<T>
        this.applicationCache.set(cachedTypeReplacement, newType);
        // convert super
        if (!typeGuards_1.empty(superMap)) {
            const mappedSuperReplacement = this.mapArguments(typeArguments, superMap.mapping);
            newType.addSuper(typeCreators_1.noMap(this.applyToType(superMap.type, mappedSuperReplacement)));
        }
        // convert super
        if (!typeGuards_1.empty(indexerMap)) {
            const mappedIndxerArguments = this.mapArguments(typeArguments, indexerMap.mapping);
            newType.addIndexer(typeCreators_1.noMap(indexerMap.type.apply(mappedIndxerArguments)));
        }
        // convert suffixes
        for (const [, suffix] of suffixes) {
            const mappedSuffixReplacement = this.mapArguments(typeArguments, suffix.mapping);
            newType.addSuffixes(typeCreators_1.noMap(this.applyToType(suffix.type, mappedSuffixReplacement)));
        }
        // convert coercions
        newType.addCoercion(...Array.from(type.getCoercions()).map(c => this.applyToType(c, typeArguments)));
        // convert operators
        for (const [kind, operators] of type.getOperators()) {
            for (const operator of operators) {
                const returnType = this.applyToType(operator.returnType, typeArguments);
                const firstOperand = this.applyToType(operator.firstOperand, typeArguments);
                const secondOperand = typeGuards_1.empty(operator.secondOperand)
                    ? undefined
                    : this.applyToType(operator.secondOperand, typeArguments);
                newType.addOperators(new operator_1.Operator(firstOperand, kind, returnType, secondOperand));
            }
        }
        return newType;
    }
}
exports.TypeBinder = TypeBinder;
//# sourceMappingURL=typeBinder.js.map