import { IParametricType, Access, TypeKind, TypeMap, IParametricCallSignature, OperatorKind, IType, IParametricIndexer } from '../types';
import { Operator } from '../types/operator';
import { TypeBinder } from '../binders/typeBinder';
/**
 * A class representing a parametric type in Kerboscript
 */
export declare class ParametricType implements IParametricType {
    /**
     * What is the name of this parametric type
     */
    readonly name: string;
    /**
     * What is the access of this type
     */
    readonly access: Access;
    /**
     * What is the kind of this type
     */
    readonly kind: TypeKind;
    /**
     * Is this the any type
     */
    readonly anyType: boolean;
    /**
     * What is the call signature of this type
     */
    protected callSignature?: TypeMap<IParametricCallSignature>;
    /**
     * What is the super type of this type
     */
    protected superType?: TypeMap<IParametricType>;
    /**
     * What is the indexer of this type
     */
    protected indexer?: TypeMap<IParametricIndexer>;
    /**
     * What are the suffixes of this type
     */
    protected suffixes: Map<string, TypeMap<IParametricType>>;
    /**
     * What are the operators of this type
     */
    protected operators: Map<OperatorKind, Operator<IParametricType>[]>;
    /**
     * The type binder
     */
    protected binder: TypeBinder;
    /**
     * What are the coercible type
     */
    protected coercibleTypes: Set<IParametricType>;
    /**
     * Construct a new parametric type
     * @param name name of this type
     * @param access what is the access of this type
     * @param typeParameters what type parameters are present in type
     * @param kind the type kind
     */
    constructor(name: string, access: Access, typeParameters: string[], kind: TypeKind);
    /**
     * Add a call signature to this type
     * @param callSignature call signature
     */
    addCallSignature(callSignature: TypeMap<IParametricCallSignature>): void;
    /**
     * Add a super type to this type
     * @param type super type to add
     */
    addSuper(type: TypeMap<IParametricType>): void;
    addIndexer(indexer: TypeMap<IParametricIndexer>): void;
    /**
     * Check that the type map is applicable to this type
     * @param typeMap type map
     */
    private checkMapping;
    /**
     * Add coercions to this type
     * @param types coercions to add
     */
    addCoercion(...types: IParametricType[]): void;
    /**
     * Add suffixes to this type
     * @param suffixes suffixes to add
     */
    addSuffixes(...suffixes: TypeMap<IParametricType>[]): void;
    /**
     * Add operators to this type
     * @param operators operators to add
     */
    addOperators(...operators: Operator<IParametricType>[]): void;
    /**
     * Is this type a subtype of another type
     * @param type type to check against
     */
    isSubtypeOf(type: IParametricType): boolean;
    /**
     * Can this type be coerced from the provided type
     * @param type to coerce from
     */
    canCoerceFrom(type: IParametricType): boolean;
    /**
     * Get the type parameters
     */
    getTypeParameters(): IParametricType[];
    /**
     * Get the super type of this type
     */
    getSuperType(): IParametricType | undefined;
    /**
     * Get the call signature of this type
     */
    getCallSignature(): Maybe<IParametricCallSignature>;
    /**
     * Get the indexer of this type
     */
    getIndexer(): Maybe<IParametricIndexer>;
    /**
     * Get the coercions of this type
     */
    getCoercions(): Set<IParametricType>;
    /**
     * Get the suffixes of this type
     */
    getSuffixes(): Map<string, IParametricType>;
    /**
     * Get an operator for the provided kind
     * @param kind operator kind
     * @param other other type if binary
     */
    getOperator(kind: OperatorKind, other?: IParametricType): Maybe<Operator<IParametricType>>;
    /**
     * Get all operators
     */
    getOperators(): Map<OperatorKind, Operator<IParametricType>[]>;
    /**
     * Get a string representing of this type
     */
    toString(): string;
    apply(typeArguments: Map<IParametricType, IType> | IType): IType;
}
