"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../types");
const typeBinder_1 = require("../binders/typeBinder");
const typeGuards_1 = require("../../utilities/typeGuards");
/**
 * A class representing a parametric type in Kerboscript
 */
class ParametricType {
    /**
     * Construct a new parametric type
     * @param name name of this type
     * @param access what is the access of this type
     * @param typeParameters what type parameters are present in type
     * @param kind the type kind
     */
    constructor(name, access, typeParameters, kind) {
        this.name = name;
        this.access = access;
        this.kind = kind;
        this.anyType = false;
        this.binder = new typeBinder_1.TypeBinder(typeParameters);
        this.superType = undefined;
        this.indexer = undefined;
        this.suffixes = new Map();
        this.coercibleTypes = new Set();
        this.operators = new Map();
    }
    /**
     * Add a call signature to this type
     * @param callSignature call signature
     */
    addCallSignature(callSignature) {
        this.checkMapping(callSignature);
        this.callSignature = callSignature;
    }
    /**
     * Add a super type to this type
     * @param type super type to add
     */
    addSuper(type) {
        if (!typeGuards_1.empty(this.superType)) {
            throw new Error(`Super type for ${this.name} has already been set.`);
        }
        // check if super has type parameters
        this.checkMapping(type);
        this.superType = type;
    }
    addIndexer(indexer) {
        if (!typeGuards_1.empty(this.indexer)) {
            throw new Error(`Indexer for ${this.name} has already been set.`);
        }
        this.checkMapping(indexer);
        this.indexer = indexer;
    }
    /**
     * Check that the type map is applicable to this type
     * @param typeMap type map
     */
    checkMapping(typeMap) {
        const { type, mapping } = typeMap;
        // check if super has type parameters
        if (type.getTypeParameters().length > 0) {
            const targetTypeParams = type.getTypeParameters();
            const thisTypeParams = this.getTypeParameters();
            // check length
            if (mapping.size !== targetTypeParams.length) {
                throw new Error(`Type has type parameters ${targetTypeParams.join(', ')}` +
                    ` but was only given ${mapping.size} arguments`);
            }
            // check matching
            for (const [key, value] of mapping) {
                if (!thisTypeParams.includes(key)) {
                    throw new Error(`Type ${this.name} does not have a type parameter ${key.name}`);
                }
                if (!targetTypeParams.includes(value)) {
                    throw new Error(`Type ${type.name} does not have a type parameter ${value.name}`);
                }
            }
        }
    }
    /**
     * Add coercions to this type
     * @param types coercions to add
     */
    addCoercion(...types) {
        for (const type of types) {
            if (this.coercibleTypes.has(type)) {
                throw new Error(`Coercible type ${type.name} has already been added to ${this.name}`);
            }
            this.coercibleTypes.add(type);
        }
    }
    /**
     * Add suffixes to this type
     * @param suffixes suffixes to add
     */
    addSuffixes(...suffixes) {
        for (const suffix of suffixes) {
            if (this.suffixes.has(suffix.type.name)) {
                throw new Error(`Duplicate suffix of ${suffix.type.name} added to ${this.name}`);
            }
            // check mapping
            this.checkMapping(suffix);
            this.suffixes.set(suffix.type.name, suffix);
        }
    }
    /**
     * Add operators to this type
     * @param operators operators to add
     */
    addOperators(...operators) {
        for (const operator of operators) {
            if (!this.operators.has(operator.operator)) {
                this.operators.set(operator.operator, []);
            }
            const operatorsKind = this.operators.get(operator.operator);
            // should never happen
            if (typeGuards_1.empty(operatorsKind)) {
                throw new Error(`Operator kind ${types_1.OperatorKind[operator.operator]} not found for ${this.name}`);
            }
            // check to make sure we didn't add two operators with the same
            // other operand
            for (const existingOperator of operatorsKind) {
                if (existingOperator.secondOperand === operator.secondOperand) {
                    let message;
                    const { secondOperand: otherOperand } = existingOperator;
                    if (typeGuards_1.empty(otherOperand)) {
                        message =
                            `Operator of kind ${types_1.OperatorKind[operator.operator]}` +
                                ` already exists between for ${this.name}`;
                    }
                    else {
                        message =
                            `Operator of kind ${types_1.OperatorKind[operator.operator]} already exists between` +
                                ` ${this.name} and ${otherOperand.name}`;
                    }
                    throw new Error(message);
                }
            }
            operatorsKind.push(operator);
        }
    }
    /**
     * Is this type a subtype of another type
     * @param type type to check against
     */
    isSubtypeOf(type) {
        if (type === this) {
            return true;
        }
        return typeGuards_1.empty(this.superType)
            ? false
            : this.superType.type.isSubtypeOf(type);
    }
    /**
     * Can this type be coerced from the provided type
     * @param type to coerce from
     */
    canCoerceFrom(type) {
        // if type no coercion needed
        if (type === this) {
            return true;
        }
        // if any type can coerce
        if (type.anyType) {
            return true;
        }
        // Are we directly a type that can be coerced
        if (this.coercibleTypes.has(type)) {
            return true;
        }
        // if we're a subtype of this type
        if (type.isSubtypeOf(this)) {
            return true;
        }
        // Are we a subtype of one of the coercible types
        for (const coercibleType of this.coercibleTypes) {
            if (type.isSubtypeOf(coercibleType)) {
                return true;
            }
        }
        return false;
    }
    /**
     * Get the type parameters
     */
    getTypeParameters() {
        return [...this.binder.typeParameters];
    }
    /**
     * Get the super type of this type
     */
    getSuperType() {
        return this.superType && this.superType.type;
    }
    /**
     * Get the call signature of this type
     */
    getCallSignature() {
        return this.callSignature && this.callSignature.type;
    }
    /**
     * Get the indexer of this type
     */
    getIndexer() {
        return this.indexer && this.indexer.type;
    }
    /**
     * Get the coercions of this type
     */
    getCoercions() {
        return this.coercibleTypes;
    }
    /**
     * Get the suffixes of this type
     */
    getSuffixes() {
        const suffixes = new Map();
        for (const [name, typeMap] of this.suffixes) {
            suffixes.set(name, typeMap.type);
        }
        if (!typeGuards_1.empty(this.superType)) {
            for (const [name, value] of this.superType.type.getSuffixes()) {
                suffixes.set(name, value);
            }
        }
        return suffixes;
    }
    /**
     * Get an operator for the provided kind
     * @param kind operator kind
     * @param other other type if binary
     */
    getOperator(kind, other) {
        const operators = this.operators.get(kind);
        if (typeGuards_1.empty(operators)) {
            return operators;
        }
        for (const operator of operators) {
            // check if operator is unary other see if it can coerced into other operand
            if (typeGuards_1.empty(operator.secondOperand)) {
                if (typeGuards_1.empty(other)) {
                    return operator;
                }
            }
            else if (!typeGuards_1.empty(other) && operator.secondOperand.canCoerceFrom(other)) {
                return operator;
            }
        }
        return undefined;
    }
    /**
     * Get all operators
     */
    getOperators() {
        return this.operators;
    }
    /**
     * Get a string representing of this type
     */
    toString() {
        const typeParameters = this.getTypeParameters();
        const typeParameterStr = typeParameters.length > 0
            ? `<${typeParameters.map(t => t.toString()).join(', ')}>`
            : '';
        // no call signature just return type name
        if (typeGuards_1.empty(this.callSignature)) {
            return `${this.name}${typeParameterStr}`;
        }
        // if it's gettable or settable then show return type
        if (this.access.get || this.access.set) {
            return `${typeParameterStr}${this.callSignature.type
                .returns()
                .toString()}`;
        }
        // else full call signature
        return `${typeParameterStr}${this.callSignature.type.toString()}`;
    }
    apply(typeArguments) {
        if (typeArguments instanceof Map) {
            return this.binder.apply(this, typeArguments, this.suffixes, this.callSignature, this.superType, this.indexer);
        }
        const typeParameters = this.getTypeParameters();
        if (typeParameters.length !== 1) {
            throw new Error('Must provide a type map if more than one parameter is present.');
        }
        return this.binder.apply(this, new Map([[typeParameters[0], typeArguments]]), this.suffixes, this.callSignature, this.superType, this.indexer);
    }
}
exports.ParametricType = ParametricType;
//# sourceMappingURL=parametricType.js.map