"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const SuffixTerm = __importStar(require("../parser/suffixTerm"));
const Expr = __importStar(require("../parser/expr"));
const Stmt = __importStar(require("../parser/stmt"));
const types_1 = require("./types");
const logger_1 = require("../utilities/logger");
const typeGuards_1 = require("../utilities/typeGuards");
const structure_1 = require("./ksTypes/primitives/structure");
const constants_1 = require("../utilities/constants");
const tokentypes_1 = require("../entities/tokentypes");
const node_1 = require("./ksTypes/node");
const typeCreators_1 = require("./typeCreators");
const arrayUtils_1 = require("../utilities/arrayUtils");
const typeUtilities_1 = require("./typeUtilities");
const void_1 = require("./ksTypes/primitives/void");
const boolean_1 = require("./ksTypes/primitives/boolean");
const string_1 = require("./ksTypes/primitives/string");
const scalar_1 = require("./ksTypes/primitives/scalar");
const typeHelpers_1 = require("./typeHelpers");
const delegate_1 = require("./ksTypes/primitives/delegate");
const typeNode_1 = require("./typeNode");
const types_2 = require("../analysis/types");
const positionUtils_1 = require("../utilities/positionUtils");
const list_1 = require("./ksTypes/collections/list");
const bodyTarget_1 = require("./ksTypes/orbital/bodyTarget");
const vesselTarget_1 = require("./ksTypes/orbital/vesselTarget");
const volume_1 = require("./ksTypes/io/volume");
const volumeItem_1 = require("./ksTypes/io/volumeItem");
const part_1 = require("./ksTypes/parts/part");
const path_1 = require("./ksTypes/io/path");
const vscode_languageserver_1 = require("vscode-languageserver");
const diagnosticsUtils_1 = require("../utilities/diagnosticsUtils");
const suffixTypeNode_1 = require("./suffixTypeNode");
const engine_1 = require("./ksTypes/parts/engine");
const dockingPort_1 = require("./ksTypes/parts/dockingPort");
const vesselSensors_1 = require("./ksTypes/vessel/vesselSensors");
const kosProcessorFields_1 = require("./ksTypes/kosProcessorFields");
const element_1 = require("./ksTypes/parts/element");
const aggregateResource_1 = require("./ksTypes/parts/aggregateResource");
const variadicType_1 = require("./types/variadicType");
/**
 * The type checker attempts to identify places where the rules of the type
 * system are broken.
 */
class TypeChecker {
    constructor(script, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        /**
         * The cached bound check statement
         */
        this.checkStmtBind = this.checkStmt.bind(this);
        this.script = script;
        this.logger = logger;
        this.tracer = tracer;
    }
    /**
     * Check the source file for type errors
     */
    check() {
        // resolve the sequence of statements
        try {
            const splits = this.script.uri.split('/');
            const file = splits[splits.length - 1];
            this.logger.info(`Type Checking started for ${file}.`);
            const typeErrors = this.checkStmts(this.script.stmts);
            this.logger.info(`Type Checking finished for ${file}`);
            if (typeErrors.length) {
                this.logger.warn(`Type Checking encountered ${typeErrors.length} errors`);
            }
            return typeErrors;
        }
        catch (err) {
            this.logger.error('Error occurred in type checker');
            logger_1.logException(this.logger, this.tracer, err, 4 /* error */);
            return [];
        }
    }
    /**
     * check a collection of statements
     * @param stmts statements to check
     */
    checkStmts(stmts) {
        return accumulateErrors(stmts, this.checkStmtBind);
    }
    /**
     * Check an statement for errors
     * @param stmt statement to check
     */
    checkStmt(stmt) {
        return stmt.accept(this);
    }
    /**
     * Check an expression for errors
     * @param expr expression to check
     */
    checkExpr(expr) {
        return expr.accept(this);
    }
    /**
     * Check a suffix term for errors
     * @param suffixTerm suffix term to check
     * @param builder type resolved so far
     */
    checkSuffixTerm(suffixTerm, builder) {
        return suffixTerm.acceptParam(this, builder);
    }
    // ----------------------------- Declaration -----------------------------------------
    /**
     * Visit a variable declaration
     * @param decl variable declaration
     */
    visitDeclVariable(decl) {
        const result = this.checkExpr(decl.value);
        const { tracker } = decl.identifier;
        if (this.isBasicTracker(tracker)) {
            tracker.declareType(result.type);
        }
        return result.errors;
    }
    /**
     * Vist a lock declaration
     * @param decl lock declaration
     */
    visitDeclLock(decl) {
        const result = this.checkExpr(decl.value);
        const { tracker } = decl.identifier;
        if (this.isBasicTracker(tracker)) {
            tracker.declareType(result.type);
        }
        return result.errors;
    }
    /**
     * Visit a function declaration
     * @param decl function declaration
     */
    visitDeclFunction(decl) {
        const errors = this.checkStmt(decl.block);
        const { tracker } = decl.identifier;
        if (!typeGuards_1.empty(tracker) && tracker.kind === 0 /* basic */) {
            const { symbol } = tracker.declared;
            if (symbol.tag === types_2.KsSymbolKind.function) {
                const paramsTypes = [];
                // TODO eventually we should tag ks parameter to the function type
                for (let i = 0; i < symbol.requiredParameters; i += 1) {
                    paramsTypes.push(structure_1.structureType);
                }
                for (let i = 0; i < symbol.optionalParameters; i += 1) {
                    paramsTypes.push(structure_1.structureType);
                }
                const returnType = symbol.returnValue ? structure_1.structureType : void_1.voidType;
                const funcType = typeCreators_1.createFunctionType(tracker.declared.symbol.name.lookup, returnType, ...paramsTypes);
                tracker.declareType(funcType);
            }
        }
        return errors;
    }
    /**
     * Visit a parameter declaration
     * @param decl parameter declaration
     */
    visitDeclParameter(decl) {
        let errors = [];
        // loop over normal parameters
        for (const required of decl.requiredParameters) {
            const { tracker } = required.identifier;
            if (this.isBasicTracker(tracker)) {
                tracker.declareType(structure_1.structureType);
            }
        }
        // loop over defaulted parameters
        for (const optional of decl.optionalParameters) {
            const valueResult = this.checkExpr(optional.value);
            const { tracker } = optional.identifier;
            if (this.isBasicTracker(tracker)) {
                tracker.declareType(valueResult.type);
            }
            errors = errors.concat(valueResult.errors);
        }
        return errors;
    }
    // ----------------------------- Statements -----------------------------------------
    /**
     * Vist an invalid statement
     * @param stmt invalid statement
     */
    visitStmtInvalid(stmt) {
        if (typeGuards_1.empty(stmt.partial)) {
            return [];
        }
        const errors = [];
        // check parsed partial nodes
        for (const node of Object.values(stmt.partial)) {
            if (node instanceof Stmt.Stmt) {
                errors.push(...this.checkStmt(node));
            }
            if (node instanceof Expr.Expr) {
                const result = this.checkExpr(node);
                errors.push(...result.errors);
            }
        }
        return errors;
    }
    /**
     * Vist a block statement
     * @param stmt statement block
     */
    visitBlock(stmt) {
        return accumulateErrors(stmt.stmts, this.checkStmtBind);
    }
    /**
     * Visit an statement expression
     * @param stmt statement expression
     */
    visitExpr(stmt) {
        const result = this.checkExpr(stmt.suffix);
        return result.errors;
    }
    /**
     * Visit an on off statement
     * @param stmt on / off statement
     */
    visitOnOff(stmt) {
        const result = this.checkExpr(stmt.suffix);
        return result.errors;
    }
    /**
     * Visit a command statement
     * @param _ command statement
     */
    visitCommand(_) {
        return [];
    }
    /**
     * Visit a command expression statement
     * @param stmt command expression statement
     */
    visitCommandExpr(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        switch (stmt.command.type) {
            // commands for adding and removing nodes
            case tokentypes_1.TokenType.add:
            case tokentypes_1.TokenType.remove:
                // expression must be a node type for node commands
                if (!node_1.nodeType.canCoerceFrom(result.type)) {
                    const command = stmt.command.type === tokentypes_1.TokenType.add ? 'add' : 'remove';
                    errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, `${command} expected a node.` +
                        ' Node may not able to be  be coerced into node type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                break;
            // command to edit a file
            case tokentypes_1.TokenType.edit:
                if (!node_1.nodeType.canCoerceFrom(result.type)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Path may not be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                break;
            default:
                throw new Error('Unexpected token type found in command expression');
        }
        return result.errors;
    }
    /**
     * Visit an unset statement
     * @param _ unset statement
     */
    visitUnset(_) {
        return [];
    }
    /**
     * Visit an unlock statement
     * @param _ unlock statement
     */
    visitUnlock(_) {
        return [];
    }
    // visit set
    /**
     * Visit set statement
     * @param stmt set statement
     */
    visitSet(stmt) {
        const exprResult = this.checkExpr(stmt.value);
        const errors = exprResult.errors;
        const suffixResult = this.checkExpr(stmt.suffix);
        errors.push(...suffixResult.errors);
        // check if suffix is settable
        if (!stmt.suffix.isSettable()) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set ${this.nodeError(stmt.suffix)} as it is a call`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        const { atom, trailers } = stmt.suffix.suffixTerm;
        // if a suffix trailer exists we are a full suffix
        if (!typeGuards_1.empty(stmt.suffix.trailer) || trailers.length > 0) {
            if (!suffixResult.type.canCoerceFrom(exprResult.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set suffix ${this.nodeError(stmt.suffix)}` +
                    `of type ${suffixResult.type.name} to ${exprResult.type.name}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
            return errors;
        }
        if (atom instanceof SuffixTerm.Identifier) {
            const { tracker } = atom.token;
            if (this.isBasicTracker(tracker)) {
                // update declare or set type
                if (tracker.declared.symbol.name === atom.token) {
                    tracker.declareType(exprResult.type);
                }
                else {
                    tracker.setType(atom.token, exprResult.type);
                }
            }
            else {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.value, 'TODO visit set, this should not occur', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        else {
            // was not found to be a valid target of setting
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set ${stmt.suffix.toString()}, must be identifier, or suffix`, vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit lazy global directive
    visitLazyGlobal(_) {
        return [];
    }
    // visit if statement
    visitIf(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!boolean_1.booleanType.canCoerceFrom(conditionResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be  be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return typeGuards_1.empty(stmt.elseStmt)
            ? errors.concat(this.checkStmt(stmt.body))
            : errors.concat(this.checkStmt(stmt.body), this.checkStmt(stmt.elseStmt));
    }
    // visit else statement
    visitElse(stmt) {
        return this.checkStmt(stmt.body);
    }
    // visit until statement
    visitUntil(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!boolean_1.booleanType.canCoerceFrom(conditionResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    // visit from loop
    visitFrom(stmt) {
        let errors = this.checkStmt(stmt.initializer);
        const conditionResult = this.checkExpr(stmt.condition);
        errors = errors.concat(conditionResult.errors);
        if (!boolean_1.booleanType.canCoerceFrom(conditionResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.increment), this.checkStmt(stmt.body));
    }
    // vist when statment
    visitWhen(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!boolean_1.booleanType.canCoerceFrom(conditionResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    // visit return
    visitReturn(stmt) {
        const errors = [];
        if (!typeGuards_1.empty(stmt.value)) {
            // TODO maybe update function type?
        }
        return errors;
    }
    // visit break
    visitBreak(_) {
        return [];
    }
    // visit switch
    visitSwitch(stmt) {
        const result = this.checkExpr(stmt.target);
        let errors = result.errors;
        if (!string_1.stringType.canCoerceFrom(result.type)) {
            errors = errors.concat(diagnosticsUtils_1.createDiagnostic(stmt.target, 'May not be a string identifer for volume', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    /**
     * Visit a for loop
     * @param stmt for loop statement
     */
    visitFor(stmt) {
        const result = this.checkExpr(stmt.collection);
        let errors = [];
        const { type } = result;
        if (type.kind !== 0 /* basic */ || !type.getSuffixes().has(constants_1.iterator)) {
            errors = errors.concat(diagnosticsUtils_1.createDiagnostic(stmt.collection, 'May not be a valid enumerable type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        // TODO may be able to detect if type is really pure and not mixed
        const { tracker } = stmt.element;
        if (this.isBasicTracker(tracker)) {
            const collectionIterator = type.getSuffixes().get(constants_1.iterator);
            if (!typeGuards_1.empty(collectionIterator)) {
                const enumerator = collectionIterator.getAssignmentType();
                const value = enumerator.getSuffixes().get('value');
                const setType = (value && value.getAssignmentType()) || structure_1.structureType;
                tracker.declareType(setType);
            }
            else {
                tracker.declareType(structure_1.structureType);
            }
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    /**
     * Visit on statement
     * @param stmt on statement
     */
    visitOn(stmt) {
        const result = this.checkExpr(stmt.suffix);
        const errors = [];
        if (!boolean_1.booleanType.canCoerceFrom(result.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    /**
     * Visit toggle statement
     * @param stmt toggle statement
     */
    visitToggle(stmt) {
        const result = this.checkExpr(stmt.suffix);
        const errors = result.errors;
        if (!boolean_1.booleanType.canCoerceFrom(result.type)) {
            // can only toggle boolean values
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, 'Toggle requires a boolean type. ' +
                'This may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return result.errors;
    }
    /**
     * Visit wait statement
     * @param stmt wait statement
     */
    visitWait(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        if (typeGuards_1.empty(stmt.until)) {
            // no until wait a set amount of time
            if (!scalar_1.scalarType.canCoerceFrom(result.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Wait requires a scalar type. ' +
                    'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        else {
            // wait until condition
            if (!boolean_1.booleanType.canCoerceFrom(result.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Wait requires a boolean type. ' +
                    'This may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        return errors;
    }
    /**
     * Visit log statement
     * @param stmt log statement
     */
    visitLog(stmt) {
        const exprResult = this.checkExpr(stmt.expr);
        const logResult = this.checkExpr(stmt.target);
        const errors = exprResult.errors.concat(logResult.errors);
        if (!string_1.stringType.canCoerceFrom(exprResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Can only log a string type. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!string_1.stringType.canCoerceFrom(exprResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Can only log to a path. ', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit copy
    visitCopy(stmt) {
        const sourceResult = this.checkExpr(stmt.target);
        const targetResult = this.checkExpr(stmt.destination);
        const errors = sourceResult.errors.concat(targetResult.errors);
        if (!string_1.stringType.canCoerceFrom(sourceResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only copy from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!string_1.stringType.canCoerceFrom(sourceResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.destination, 'Can only copy to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit rename
    visitRename(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const alternativeResult = this.checkExpr(stmt.alternative);
        const errors = targetResult.errors.concat(alternativeResult.errors);
        if (!string_1.stringType.canCoerceFrom(targetResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only rename from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!string_1.stringType.canCoerceFrom(targetResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.alternative, 'Can only rename to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    visitDelete(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const errors = targetResult.errors;
        if (!string_1.stringType.canCoerceFrom(targetResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only delete from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (typeGuards_1.empty(stmt.volume)) {
            return errors;
        }
        const volumeResult = this.checkExpr(stmt.volume);
        if (!string_1.stringType.canCoerceFrom(targetResult.type) &&
            !path_1.pathType.canCoerceFrom(targetResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.volume, 'Can only rename to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(volumeResult.errors);
    }
    visitRun(_) {
        return [];
    }
    visitRunPath(_) {
        return [];
    }
    visitRunPathOnce(_) {
        return [];
    }
    visitCompile(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const errors = targetResult.errors;
        if (!string_1.stringType.canCoerceFrom(targetResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only compile from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (typeGuards_1.empty(stmt.destination)) {
            return errors;
        }
        const destinationResult = this.checkExpr(stmt.destination);
        if (!string_1.stringType.canCoerceFrom(destinationResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.destination, 'Can only compile to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(destinationResult.errors);
    }
    /**
     * Visit list statement
     * @param stmt list statement
     */
    visitList(stmt) {
        const { target, collection } = stmt;
        if (typeGuards_1.empty(target) || typeGuards_1.empty(collection)) {
            return [];
        }
        let finalType;
        const errors = [];
        // determine the type that list returns
        switch (collection.lookup) {
            case 'bodies':
                finalType = bodyTarget_1.bodyTargetType;
                break;
            case 'targets':
                finalType = vesselTarget_1.vesselTargetType;
                break;
            case 'elements':
                finalType = element_1.elementType;
                break;
            case 'resources':
                finalType = list_1.listType.apply(aggregateResource_1.aggregateResourceType);
                break;
            case 'parts':
                finalType = part_1.partType;
                break;
            case 'sensors':
                finalType = vesselSensors_1.vesselSensorsType;
                break;
            case 'dockingports':
                finalType = dockingPort_1.dockingPortType;
                break;
            case 'engines':
                finalType = engine_1.engineType;
                break;
            case 'files':
                finalType = volumeItem_1.volumeItemType;
                break;
            case 'fonts':
                finalType = string_1.stringType;
                break;
            case 'volumes':
                finalType = volume_1.volumeType;
                break;
            case 'processors':
                finalType = kosProcessorFields_1.kosProcessorFieldsType;
                break;
            default:
                finalType = structure_1.structureType;
                errors.push(diagnosticsUtils_1.createDiagnostic(collection, 'Not a valid list identifier', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        const { tracker } = target;
        if (this.isBasicTracker(tracker)) {
            tracker.setType(target, list_1.listType.apply(finalType));
        }
        return errors;
    }
    // visit empty statement
    visitEmpty(_) {
        return [];
    }
    // vist print statement
    visitPrint(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        if (!structure_1.structureType.canCoerceFrom(result.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Cannot print a function, can only print structures', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit invalid expression
    visitExprInvalid(_) {
        return { type: structure_1.structureType, errors: [] };
    }
    // ----------------------------- Expressions -----------------------------------------
    /**
     * Check the ternary expression TODO need union type
     * @param expr ternary expression
     */
    visitTernary(expr) {
        const conditionResult = this.checkExpr(expr.condition);
        const trueResult = this.checkExpr(expr.trueBranch);
        const falseResult = this.checkExpr(expr.falseBranch);
        const errors = conditionResult.errors;
        errors.push(...trueResult.errors, ...falseResult.errors);
        if (!boolean_1.booleanType.canCoerceFrom(conditionResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(expr.condition, 'condition must be able to be coerced into boolean', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return { errors, type: trueResult.type };
    }
    /**
     * Visit binary expression of form `a op b`
     * @param expr binary expression
     */
    visitBinary(expr) {
        const rightResult = this.checkExpr(expr.right);
        const leftResult = this.checkExpr(expr.left);
        const operatorKind = typeUtilities_1.binaryOperatorMap.get(expr.operator.type);
        if (!typeGuards_1.empty(operatorKind)) {
            switch (operatorKind) {
                case types_1.OperatorKind.and:
                case types_1.OperatorKind.or:
                    return this.checkLogical(expr, leftResult, rightResult);
                default:
                    return this.checkBinary(expr, leftResult, rightResult, operatorKind);
            }
        }
        throw new Error(`Unexpected token ${expr.operator.typeString} type encountered in binary expression`);
    }
    /**
     * Visit unary expression of form `op a`
     * @param expr unary expression
     */
    visitUnary(expr) {
        const result = this.checkExpr(expr.factor);
        const errors = result.errors;
        switch (expr.operator.type) {
            case tokentypes_1.TokenType.defined:
                return {
                    errors,
                    type: boolean_1.booleanType,
                };
            case tokentypes_1.TokenType.not:
                if (!boolean_1.booleanType.canCoerceFrom(result.type)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(expr.factor, 'Can only apply not operator to booleans', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                return {
                    errors,
                    type: boolean_1.booleanType,
                };
            case tokentypes_1.TokenType.minus:
            case tokentypes_1.TokenType.plus:
                const operatorKind = typeUtilities_1.unaryOperatorMap.get(expr.operator.type);
                if (!typeGuards_1.empty(operatorKind)) {
                    return this.checkUnary(expr, result, operatorKind);
                }
        }
        throw new Error(`Invalid Token ${expr.operator.typeString} for unary operator.`);
    }
    /**
     * Visit a factor expression of form `a ^ b`.
     * @param expr factor expression
     */
    visitFactor(expr) {
        const suffixResult = this.checkExpr(expr.suffix);
        const exponentResult = this.checkExpr(expr.exponent);
        if (expr.power.type !== tokentypes_1.TokenType.power) {
            throw new Error('Factor does not contain power operator');
        }
        return this.checkBinary(expr, suffixResult, exponentResult, types_1.OperatorKind.power);
    }
    /**
     * Visit a suffix
     * @param expr suffix expression
     */
    visitSuffix(expr) {
        const { suffixTerm, trailer } = expr;
        const [firstTrailer, ...remainingTrailers] = suffixTerm.trailers;
        const builder = new suffixTypeNode_1.SuffixTypeBuilder();
        const errors = this.checkSuffixTerm(suffixTerm.atom, builder);
        if (!typeGuards_1.empty(firstTrailer)) {
            // handle case were suffix is actually a function call
            if (firstTrailer instanceof SuffixTerm.Call) {
                const tracker = this.atomTracker(suffixTerm.atom);
                errors.push(...this.visitFunctionCall(firstTrailer, tracker, builder));
            }
            else {
                errors.push(...this.checkSuffixTerm(firstTrailer, builder));
            }
            // handle remaining suffix term trailers
            for (const trailer of remainingTrailers) {
                errors.push(...this.checkSuffixTerm(trailer, builder));
            }
        }
        // if we have a trailer check that as well
        if (!typeGuards_1.empty(trailer)) {
            errors.push(...this.checkSuffixTerm(trailer, builder));
        }
        return this.resultExpr(this.builderResult(builder), errors);
    }
    visitLambda(expr) {
        const errors = this.checkStmt(expr.block);
        return this.resultExpr(delegate_1.delegateType, errors);
    }
    // ----------------------------- Suffix -----------------------------------------
    visitSuffixTrailer(suffixTerm, builder) {
        // check suffix term and trailers
        const errors = this.checkSuffixTerm(suffixTerm.suffixTerm, builder);
        // if no trailer exist attempt to return
        if (!typeGuards_1.empty(suffixTerm.trailer)) {
            errors.push(...this.checkSuffixTerm(suffixTerm.trailer, builder));
        }
        return errors;
    }
    visitSuffixTermInvalid(suffixTerm, builder) {
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, suffixTerm));
        return [];
    }
    visitSuffixTerm(suffixTerm, builder) {
        // check the atom
        const errors = this.checkSuffixTerm(suffixTerm.atom, builder);
        // add any errors from trailers
        for (const trailer of suffixTerm.trailers) {
            errors.push(...this.checkSuffixTerm(trailer, builder));
        }
        return errors;
    }
    /**
     * Visit a call expression and check for type errors
     * @param call the current call expression
     * @param builder current resolved suffix expression
     */
    visitCall(call, builder) {
        if (!builder.isTrailer()) {
            throw new Error('Builder must be a trailer to be in visitCall');
        }
        const type = builder.current();
        const errors = [];
        const callSignature = type.getCallSignature();
        if (typeGuards_1.empty(callSignature)) {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, call));
            call.open.tracker = typeHelpers_1.suffixError.getTracker();
            call.close.tracker = typeHelpers_1.suffixError.getTracker();
            errors.push(diagnosticsUtils_1.createDiagnostic(call, `${type.name} has no call signature`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        builder.nodes.push(new typeNode_1.TypeNode(type, call));
        call.open.tracker = type.getTracker();
        call.close.tracker = type.getTracker();
        const params = callSignature.params();
        // handle variadic
        if (params.length === 1 && params[0].kind === 1 /* variadic */) {
            return this.visitVariadicCall(params[0], call);
        }
        // handle normal functions
        return this.visitNormalCall(params, call);
    }
    /**
     * Visit a function call site
     * @param call call suffix term expression
     * @param builder suffix type builder
     */
    visitFunctionCall(call, tracker, builder) {
        const type = builder.current();
        const errors = [];
        const callSignature = type.getCallSignature();
        // check if previous identifier resolves to a function
        // TODO figure out function trackers
        if (typeGuards_1.empty(callSignature)) {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.functionError, call));
            call.open.tracker = typeHelpers_1.functionError.getTracker();
            call.close.tracker = typeHelpers_1.functionError.getTracker();
            errors.push(diagnosticsUtils_1.createDiagnostic(call, `Type ${type.name} may not have a call signature`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        builder.nodes.push(new typeNode_1.TypeNode(type, call));
        call.open.tracker = tracker;
        call.close.tracker = tracker;
        const params = callSignature.params();
        // handle normal or variadic calls
        if (params.length === 1 && params[0].kind === 1 /* variadic */) {
            errors.push(...this.visitVariadicCall(params[0], call));
        }
        else {
            errors.push(...this.visitNormalCall(params, call));
        }
        return errors;
    }
    /**
     * Resolve variadic call for type errors
     * @param params parameter types
     * @param call current call expression
     */
    visitVariadicCall(params, call) {
        const errors = [];
        if (!(params instanceof variadicType_1.VariadicType)) {
            throw new Error('Expected variadic type.');
        }
        for (const arg of call.args) {
            // determine type of each argument
            const result = this.checkExpr(arg);
            errors.push(...result.errors);
            // add diagnostic if argument cannot be matched to parameter type
            if (!params.base.canCoerceFrom(result.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(arg, `Function argument could not be coerced into ${params.toString()}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        return errors;
    }
    /**
     * Check a normal call signature for type errors
     * @param params the parameter types
     * @param call the call expression
     */
    visitNormalCall(params, call) {
        const errors = [];
        for (const [arg, param] of arrayUtils_1.zip(call.args, params)) {
            // determine type of each argument
            const result = this.checkExpr(arg);
            errors.push(...result.errors);
            // add diagnostic if argument cannot be matched to parameter type
            if (!param.canCoerceFrom(result.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(arg, `Function argument could not be coerced into ${param.name}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        // check argument length
        if (call.args.length !== params.length) {
            errors.push(diagnosticsUtils_1.createDiagnostic(call.close, `Function expected ${params.length} parameters but was called with ${call.args.length} arguments`, vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    /**
     * visit an array index suffix expression.
     * @param suffixTerm the current array index
     * @param builder the current type
     */
    visitArrayIndex(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitArrayIndex");
        }
        const type = builder.current();
        const errors = [];
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.arrayIndexer, suffixTerm));
        // TODO confirm indexable types
        // Only lists are indexable with '#'
        if (!list_1.listType.canCoerceFrom(type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, 'indexing with # requires a list', vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        switch (suffixTerm.indexer.type) {
            // If index is integer we're already in good shape
            case tokentypes_1.TokenType.integer:
                return errors;
            // If index is identify check that it holds a integer
            case tokentypes_1.TokenType.identifier:
                const { tracker } = suffixTerm.indexer;
                if (this.isBasicTracker(tracker)) {
                    const type = tracker.getType(suffixTerm.indexer);
                    if (!typeGuards_1.empty(type) && scalar_1.integerType.canCoerceFrom(type)) {
                        return errors;
                    }
                }
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.indexer, `${suffixTerm.indexer.lexeme} is not a scalar type. ` +
                    'Can only use scalar to index with #', vscode_languageserver_1.DiagnosticSeverity.Hint));
                return errors;
            // All other cases are disallowed
            default:
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.indexer, 'Can only index an array with # scalars or variables', vscode_languageserver_1.DiagnosticSeverity.Hint));
                return errors;
        }
    }
    /**
     * visit an array bracket suffix expression.
     * @param suffixTerm the current array bracket expression
     * @param builder the suffix type builder
     */
    visitArrayBracket(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitArrayBracket");
        }
        const indexResult = this.checkExpr(suffixTerm.index);
        const errors = indexResult.errors;
        const type = builder.current().getAssignmentType();
        const indexer = type.getIndexer();
        // we either need to have an indexer or be the any type
        if (typeGuards_1.empty(indexer)) {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.indexerError, suffixTerm));
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, `${type.toString()} may not have indexer`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            suffixTerm.open.tracker = typeHelpers_1.indexerError.getTracker();
            suffixTerm.close.tracker = typeHelpers_1.indexerError.getTracker();
            return errors;
        }
        const callSignature = indexer.getCallSignature();
        if (typeGuards_1.empty(callSignature)) {
            throw new Error('Indexer should have filled call signature');
        }
        builder.nodes.push(new typeNode_1.TypeNode(indexer, suffixTerm));
        // check that index can be coerced into requested type
        if (!callSignature.params()[0].canCoerceFrom(indexResult.type)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.index, `${indexer.toString()} may indexed with of type ${callSignature.params()[0]}` + ` ${suffixTerm.toString()} is of type ${indexResult.type}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        // assign trackers
        suffixTerm.open.tracker = indexer.getTracker();
        suffixTerm.close.tracker = indexer.getTracker();
        return errors;
    }
    /**
     * visit the suffix term for delgates. This will return a new delgate type
     * @param suffixTerm the current delgate node
     * @param builder the suffix type builder
     */
    visitDelegate(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitDelegate");
        }
        const type = builder.current();
        const errors = [];
        if (type.kind !== 4 /* function */) {
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, 'Can only create delegate of functions', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.delegateCreation, suffixTerm));
        suffixTerm.atSign.tracker = typeHelpers_1.delegateCreation.getTracker();
        return errors;
    }
    /**
     * visit the suffix term for literals.
     * @param suffixTerm literal syntax node
     * @param builder the suffix type builder
     */
    visitLiteral(suffixTerm, builder) {
        if (builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitLiteral");
        }
        // if we're at the base push the new type node onto the builder
        switch (suffixTerm.token.type) {
            case tokentypes_1.TokenType.true:
            case tokentypes_1.TokenType.false:
                builder.nodes.push(new typeNode_1.TypeNode(boolean_1.booleanType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.integer:
                builder.nodes.push(new typeNode_1.TypeNode(scalar_1.integerType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.double:
                builder.nodes.push(new typeNode_1.TypeNode(scalar_1.doubleType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.string:
            case tokentypes_1.TokenType.fileIdentifier:
                builder.nodes.push(new typeNode_1.TypeNode(string_1.stringType, suffixTerm));
                return [];
            default:
                throw new Error('TODO invalid literal token found visitLiteral');
        }
    }
    /**
     * visit the suffix term for identifier.
     * @param suffixTerm identifier syntax node
     * @param builder the suffix type builder
     */
    visitIdentifier(suffixTerm, builder) {
        // if we're a trailer check for suffixes
        if (builder.isTrailer()) {
            const type = builder.current();
            const suffix = this.builderResult(builder)
                .getSuffixes()
                .get(suffixTerm.token.lookup);
            // may need to pass something in about if we're in get set context
            if (!typeGuards_1.empty(suffix)) {
                // assign type tracker
                suffixTerm.token.tracker = suffix.getTracker();
                // push new node onto builder
                builder.nodes.push(new typeNode_1.TypeNode(suffix, suffixTerm));
                return [];
            }
            // assign error suffix type
            suffixTerm.token.tracker = typeHelpers_1.suffixError.getTracker();
            // add suffix error to builder
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, suffixTerm));
            // indicate suffix not found
            return [
                diagnosticsUtils_1.createDiagnostic(suffixTerm, `Unable to find suffix ${suffixTerm.token.lookup} on type ${type.toString()}`, vscode_languageserver_1.DiagnosticSeverity.Hint),
            ];
        }
        const { tracker } = suffixTerm.token;
        // make sure we have a basic tracker
        if (this.isBasicTracker(tracker)) {
            const type = tracker.getType(suffixTerm.toLocation(this.script.uri));
            // if type is found at this location add it to builder
            if (!typeGuards_1.empty(type)) {
                builder.nodes.push(new typeNode_1.TypeNode(type, suffixTerm));
                return [];
            }
            // if we can't find the type here default to structure
            builder.nodes.push(new typeNode_1.TypeNode(structure_1.structureType, suffixTerm));
            return [
                diagnosticsUtils_1.createDiagnostic(suffixTerm, `Cannot determine type for ${suffixTerm.token.lexeme}.`, vscode_languageserver_1.DiagnosticSeverity.Hint),
            ];
        }
        // in theory we should never get here
        builder.nodes.push(new typeNode_1.TypeNode(structure_1.structureType, suffixTerm));
        // no error as this is likely a resolver error
        return [];
    }
    /**
     * visit the suffix term for grouping.
     * @param suffixTerm grouping syntax node
     * @param builder the suffix type builder
     */
    visitGrouping(suffixTerm, builder) {
        if (builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitGrouping");
        }
        const { type, errors } = this.checkExpr(suffixTerm.expr);
        // push result of grouping onto builder
        builder.nodes.push(new typeNode_1.TypeNode(type, suffixTerm));
        return errors;
    }
    // ----------------------------- Helpers -----------------------------------------
    /**
     * Get the tracker from an atom
     * @param atom atom
     */
    atomTracker(atom) {
        if (atom instanceof SuffixTerm.Identifier) {
            return atom.token.tracker;
        }
        return undefined;
    }
    /**
     * Is this tracker a basic tracker
     * @param tracker tracker to inspect
     */
    isBasicTracker(tracker) {
        return !typeGuards_1.empty(tracker) && tracker.kind === 0 /* basic */;
    }
    /**
     * Check if unary operator is valid for this type
     * @param expr unary expression
     * @param subExpression the sub expression result
     * @param operatorKind the kind of operator
     */
    checkUnary(expr, subExpression, operatorKind) {
        const subType = subExpression.type;
        const subOp = subType.getOperator(operatorKind);
        const errors = subExpression.errors;
        if (!typeGuards_1.empty(subOp) && subOp.isUnary()) {
            return { errors, type: subOp.returnType };
        }
        errors.push(this.operatorError(operatorKind, expr.factor, subType));
        return {
            errors,
            type: structure_1.structureType,
        };
    }
    /**
     * Check if a logical operator is valid for these operands
     * @param expr underlying expression
     * @param leftResult the left operand result
     * @param rightResult the right operand result
     */
    checkLogical(expr, leftResult, rightResult) {
        const leftType = leftResult.type;
        const rightType = rightResult.type;
        const errors = leftResult.errors.concat(rightResult.errors);
        // check if left can be converted to a boolean
        if (!boolean_1.booleanType.canCoerceFrom(leftType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(expr.left, `${leftType.name} cannot be converted to a boolean`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return { errors, type: boolean_1.booleanType };
        }
        // check if right can be converted to a boolean
        if (!boolean_1.booleanType.canCoerceFrom(rightType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(expr.right, `${rightType.name} cannot be converted to a boolean`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return { errors, type: boolean_1.booleanType };
        }
        return { errors, type: boolean_1.booleanType };
    }
    /**
     * Check if the current operator is valid and it's resulting type
     * @param expr the operator expression
     * @param leftResult the left type
     * @param rightResult the right type
     * @param operator the operator to consider
     */
    checkBinary(expr, leftResult, rightResult, operator) {
        const leftType = leftResult.type;
        const rightType = rightResult.type;
        const errors = leftResult.errors.concat(rightResult.errors);
        let calcType = undefined;
        // determine the type to perform the operation on
        if (leftType.isSubtypeOf(scalar_1.scalarType) && rightType.isSubtypeOf(scalar_1.scalarType)) {
            calcType = scalar_1.scalarType;
        }
        else if (leftType.isSubtypeOf(string_1.stringType) ||
            rightType.isSubtypeOf(string_1.stringType)) {
            calcType = string_1.stringType;
        }
        else if (leftType.isSubtypeOf(boolean_1.booleanType) ||
            rightType.isSubtypeOf(boolean_1.booleanType)) {
            calcType = boolean_1.booleanType;
        }
        // we found a primitive type to perform the operation on
        if (!typeGuards_1.empty(calcType)) {
            // get all operators for the given kind for the calc type
            const calcOps = calcType.getOperators().get(operator);
            // if no operators present return error
            if (!typeGuards_1.empty(calcOps)) {
                const result = this.tryBinaryOperators(calcOps, rightType, errors);
                if (!typeGuards_1.empty(result)) {
                    return result;
                }
            }
            errors.push(this.operatorError(operator, expr, leftType, rightType));
            return { errors, type: structure_1.structureType };
        }
        // try to get an operator between the two types
        const leftOps = leftType.getOperator(operator, rightType);
        const rightOps = rightType.getOperator(operator, leftType);
        // no operator found return error
        if (typeGuards_1.empty(leftOps) && typeGuards_1.empty(rightOps)) {
            errors.push(this.operatorError(operator, expr, leftType, rightType));
            return {
                errors,
                type: structure_1.structureType,
            };
        }
        // return results if found
        if (!typeGuards_1.empty(leftOps)) {
            return { errors, type: leftOps.returnType };
        }
        if (!typeGuards_1.empty(rightOps)) {
            return { errors, type: rightOps.returnType };
        }
        // else error
        errors.push(this.operatorError(operator, expr, leftType, rightType));
        return { errors, type: structure_1.structureType };
    }
    /**
     * Report an error when an operator is not supported by the provided types
     * @param operatorKind operator kind
     * @param expr expression the error occurred in
     * @param leftType the left type
     * @param rightType the right type
     */
    operatorError(operatorKind, expr, leftType, rightType) {
        if (typeGuards_1.empty(rightType)) {
            return diagnosticsUtils_1.createDiagnostic(expr, `${leftType.name} may not support the ${types_1.OperatorKind[operatorKind]} operator`, vscode_languageserver_1.DiagnosticSeverity.Hint);
        }
        return diagnosticsUtils_1.createDiagnostic(expr, `${leftType.name} nor ${rightType.name} supports the ${types_1.OperatorKind[operatorKind]} operator between them`, vscode_languageserver_1.DiagnosticSeverity.Hint);
    }
    /**
     * Check if any of the available binary operators can work for the other operand
     * @param operators the available operators for a given operator type
     * @param otherType the type of the other operand
     * @param errors the current set of errors
     */
    tryBinaryOperators(operators, rightType, errors) {
        for (const operator of operators) {
            if (operator.isUnary()) {
                continue;
            }
            const { secondOperand: otherOperand, returnType } = operator;
            if (!typeGuards_1.empty(otherOperand) && otherOperand.canCoerceFrom(rightType)) {
                return { errors, type: returnType };
            }
        }
        return undefined;
    }
    /**
     * Return the resultant type with any errors
     * @param type current type
     * @param errors accumulated errors
     */
    resultExpr(type, ...errors) {
        return {
            type,
            errors: [].concat(...errors),
        };
    }
    /**
     * Get the current result of the suffix type buildcallSignatureer
     * @param builder suffix type builder
     */
    builderResult(builder) {
        const current = builder.current();
        return current.getAssignmentType();
    }
    /**
     * Convert node to string and relevant range
     * @param node node to create message for
     */
    nodeError(node) {
        return `${node.toString()} ${positionUtils_1.rangeToString(node)}`;
    }
}
exports.TypeChecker = TypeChecker;
const accumulateErrors = (items, checker) => {
    const errors = [];
    for (const item of items) {
        errors.push(...checker(item));
    }
    return errors;
};
//# sourceMappingURL=typeChecker.js.map