"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const memoize_1 = require("../utilities/memoize");
const parametricType_1 = require("./parametricTypes/parametricType");
const type_1 = require("./types/type");
const variadicType_1 = require("./types/variadicType");
const callSignature_1 = require("./types/callSignature");
const parametricCallSignature_1 = require("./parametricTypes/parametricCallSignature");
const parametricIndexer_1 = require("./parametricTypes/parametricIndexer");
const indexer_1 = require("./types/indexer");
/**
 * Create a type parameter type
 * @param name name of the placeholder
 */
exports.createTypeParameter = (name) => {
    return new parametricType_1.ParametricType(name, { get: false, set: false }, [], 5 /* typeSlot */);
};
/**
 * Generate a new basic parametric type
 * @param name name of the new parametric type
 * @param typeParameterNames names of the type parameters
 */
exports.createParametricType = (name, typeParameterNames) => {
    return new parametricType_1.ParametricType(name, { get: true, set: true }, typeParameterNames, 0 /* basic */);
};
/**
 * Create an indexer
 * @param index the index type
 * @param returns the return type
 */
exports.createIndexer = (index, returns) => {
    return new indexer_1.Indexer(new callSignature_1.CallSignature([index], returns), new Map());
};
/**
 * Create an parametric indexer
 * @param typeParameters type parameters for the indexer
 * @param index the index type
 * @param returns the return type
 */
exports.createParametricIndexer = (typeParameters, index, returns) => {
    const indexer = new parametricIndexer_1.ParametricIndexer(typeParameters);
    const callSignature = new parametricCallSignature_1.GenericCallSignature(typeParameters);
    const typePlaceholders = callSignature.getTypeParameters();
    callSignature.addParams(mapTypeWithParameters(callSignature, typePlaceholders, index));
    callSignature.addReturn(mapTypeWithParameters(callSignature, typePlaceholders, returns));
    indexer.addCallSignature(exports.mapTypes(indexer, callSignature));
    return indexer;
};
/**
 * Generate a new basic type
 * @param name name of the new type
 */
exports.createType = (name) => {
    return new type_1.Type(name, { get: true, set: true }, new Map(), 0 /* basic */);
};
/**
 * Generate a new parametric callable suffix type
 * @param name name of the suffix
 * @param returns return type of suffix call
 * @param params parameters of suffix call
 */
exports.createParametricArgSuffixType = (name, typeParameters, returns, ...params) => {
    const get = params.length === 0;
    const parametricType = new parametricType_1.ParametricType(name.toLowerCase(), { get, set: false }, typeParameters, 2 /* suffix */);
    const callSignature = new parametricCallSignature_1.GenericCallSignature(typeParameters);
    const typePlaceholders = callSignature.getTypeParameters();
    callSignature.addParams(...params.map(p => mapTypeWithParameters(callSignature, typePlaceholders, p)));
    callSignature.addReturn(mapTypeWithParameters(callSignature, typePlaceholders, returns));
    parametricType.addCallSignature(exports.mapTypes(parametricType, callSignature));
    return parametricType;
};
/**
 * Type map with string as parameter slot placeholder
 * @param parentType parent type
 * @param typeParameters the type parameters
 * @param type the type to map
 */
const mapTypeWithParameters = (parentType, typeParameters, type) => {
    if (typeof type !== 'string') {
        if (type.getTypeParameters().length === 0) {
            return exports.noMap(type);
        }
        return mapType(parentType, type);
    }
    for (const placeholder of typeParameters) {
        if (type === placeholder.name) {
            return exports.noMap(placeholder);
        }
    }
    throw new Error(`Unable to map ${type} to a type parameter`);
};
/**
 * Generate a new callable suffix type
 * @param name name of the suffix
 * @param returns return type of suffix call
 * @param params parameters of suffix call
 */
exports.createArgSuffixType = (name, returns, ...params) => {
    const get = params.length === 0;
    return new type_1.Type(name.toLowerCase(), { get, set: false }, new Map(), 2 /* suffix */, new callSignature_1.CallSignature(params, returns));
};
/**
 * Generate a new get only suffix type
 * @param name name of the suffix
 * @param returns return type of suffix
 */
exports.createSuffixType = (name, returns) => {
    return new type_1.Type(name.toLowerCase(), { get: true, set: false }, new Map(), 2 /* suffix */, new callSignature_1.CallSignature([], returns));
};
/**
 * Generate a new get and set suffix type
 * @param name name of the suffix
 * @param returns return type of suffix
 */
exports.createSetSuffixType = (name, returns) => {
    return new type_1.Type(name.toLowerCase(), { get: true, set: true }, new Map(), 2 /* suffix */, new callSignature_1.CallSignature([], returns));
};
/**
 * Generate a new callable variadic suffix type
 * @param name name of the suffix type
 * @param returns return type of suffix call
 * @param params variadic type of parameters
 */
exports.createVarSuffixType = (name, returns, params) => {
    if (params.kind !== 1 /* variadic */) {
        throw new Error('Expected variadic type.');
    }
    return new type_1.Type(name.toLowerCase(), { get: false, set: false }, new Map(), 2 /* suffix */, new callSignature_1.CallSignature([params], returns));
};
/**
 * Generate a new function type
 * @param name name of the function
 * @param returns return type of function call
 * @param params parameters of function
 */
exports.createFunctionType = (name, returns, ...params) => {
    return new type_1.Type(name.toLowerCase(), { get: false, set: false }, new Map(), 4 /* function */, new callSignature_1.CallSignature(params, returns));
};
/**
 * Generate a new variadic function type
 * @param name name of the function
 * @param returns return type of function call
 * @param params variadic type of function call parameters
 */
exports.createVarFunctionType = (name, returns, params) => {
    if (params.kind !== 1 /* variadic */) {
        throw new Error('Expected variadic type.');
    }
    return new type_1.Type(name.toLowerCase(), { get: false, set: false }, new Map(), 4 /* function */, new callSignature_1.CallSignature([params], returns));
};
/**
 * Create variadic type
 */
exports.createVarType = memoize_1.memoize((type) => {
    if (type.kind !== 0 /* basic */) {
        throw new Error('Must provide a basic type for variadic types');
    }
    return new variadicType_1.VariadicType(type);
});
/**
 * Create a type map with no parameter mapping
 * @param type type to map
 */
exports.noMap = (type) => {
    return {
        type,
        mapping: new Map(),
    };
};
/**
 * Map a types parameters from a source to a target, assumes only one parameter
 * @param source source type
 * @param target target type
 */
exports.mapTypes = (source, target) => {
    const sourceParameters = source.getTypeParameters();
    const targetParameters = target.getTypeParameters();
    if (sourceParameters.length !== 1) {
        throw new Error(`Type ${source.name} has more than 1 type parameter`);
    }
    if (targetParameters.length !== 1) {
        throw new Error(`Type ${target.name} has more than 1 type parameter`);
    }
    return {
        type: target,
        mapping: new Map([[sourceParameters[0], targetParameters[0]]]),
    };
};
/**
 * Map type parameters from a source to a target or a target type to the source's type parameters
 * @param source source type
 * @param target target type
 */
const mapType = (source, target) => {
    const sourceParameters = source.getTypeParameters();
    const targetParameters = target.getTypeParameters();
    if (sourceParameters.length !== 1) {
        throw new Error(`Type ${source.name} has more than 1 type parameter`);
    }
    if (targetParameters.length !== 1 && target.kind !== 5 /* typeSlot */) {
        throw new Error(`Type ${target.name} has more than 1 type parameter`);
    }
    if (target.kind === 5 /* typeSlot */) {
        return {
            type: target,
            mapping: new Map([[sourceParameters[0], target]]),
        };
    }
    return {
        type: target,
        mapping: new Map([[sourceParameters[0], targetParameters[0]]]),
    };
};
//# sourceMappingURL=typeCreators.js.map