"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../types");
const typeTracker_1 = require("../../analysis/typeTracker");
const suffix_1 = require("../../entities/suffix");
const typeGuards_1 = require("../../utilities/typeGuards");
/**
 * A class representing a type in Kerboscript
 */
class Type {
    /**
     * Construct a new kerboscript type
     * @param name name of the type
     * @param access what access does this type have
     * @param typeArguments what are the type arguments of this type
     * @param kind what is the kind of this type
     * @param callSignature what is the call signature of this type
     * @param typeTemplate what is the type template of this type
     * @param anyType is this the any type
     */
    constructor(name, access, typeArguments, kind, callSignature, typeTemplate, anyType = false) {
        if (kind === 1 /* variadic */) {
            throw new Error('Cannot construct variadic type.');
        }
        this.name = name;
        this.access = access;
        this.typeArguments = typeArguments;
        this.kind = kind;
        this.callSignature = callSignature;
        this.typeTemplate = typeTemplate;
        this.anyType = anyType;
        // TODO will need to actually be robust about this
        this.tracker = new typeTracker_1.TypeTracker(new suffix_1.KsSuffix(name), this);
        this.coercibleTypes = new Set();
        this.suffixes = new Map();
        this.operators = new Map();
    }
    /**
     * Add a super type to this type
     * @param typeMap type map
     */
    addSuper(typeMap) {
        if (!typeGuards_1.empty(this.superType)) {
            throw new Error(`Super type for ${this.name} has already been set.`);
        }
        this.superType = typeMap.type;
    }
    /**
     * Add an indexer to this type
     * @param indexer indexer to add
     */
    addIndexer(indexer) {
        if (!typeGuards_1.empty(this.indexer)) {
            throw new Error(`Indexer type for ${this.name} has already been set.`);
        }
        this.indexer = indexer.type;
    }
    /**
     * Add new coercions to this type
     * @param types type to new coercions
     */
    addCoercion(...types) {
        for (const type of types) {
            if (this.coercibleTypes.has(type)) {
                throw new Error(`Coercible type ${type.name} has already been added to ${this.name}`);
            }
            this.coercibleTypes.add(type);
        }
    }
    /**
     * Add new suffix to this type
     * @param suffixes suffixes to add
     */
    addSuffixes(...suffixes) {
        for (const { type } of suffixes) {
            if (this.suffixes.has(type.name)) {
                throw new Error(`Duplicate suffix of ${type.name} added to ${this.name}`);
            }
            this.suffixes.set(type.name, type);
        }
    }
    /**
     * Add new operators to this type
     * @param operators operators to add
     */
    addOperators(...operators) {
        for (const operator of operators) {
            if (!this.operators.has(operator.operator)) {
                this.operators.set(operator.operator, []);
            }
            const operatorsKind = this.operators.get(operator.operator);
            // should never happen
            if (typeGuards_1.empty(operatorsKind)) {
                throw new Error(`Operator kind ${types_1.OperatorKind[operator.operator]} not found for ${this.name}`);
            }
            // check to make sure we didn't add two operators with the same
            // other operand
            for (const existingOperator of operatorsKind) {
                if (existingOperator.secondOperand === operator.secondOperand) {
                    let message;
                    const { secondOperand: otherOperand } = existingOperator;
                    if (typeGuards_1.empty(otherOperand)) {
                        message =
                            `Operator of kind ${types_1.OperatorKind[operator.operator]}` +
                                ` already exists between for ${this.name}`;
                    }
                    else {
                        message =
                            `Operator of kind ${types_1.OperatorKind[operator.operator]} already exists between` +
                                ` ${this.name} and ${otherOperand.name}`;
                    }
                    throw new Error(message);
                }
            }
            operatorsKind.push(operator);
        }
    }
    /**
     * Is this type a subtype for another type
     * @param type the type check against
     */
    isSubtypeOf(type) {
        if (type === this) {
            return true;
        }
        // check if super type matches provided type
        if (!typeGuards_1.empty(this.superType)) {
            const isSubtype = this.superType.isSubtypeOf(type);
            if (isSubtype) {
                return isSubtype;
            }
        }
        // check if type template matches provided type
        if (!typeGuards_1.empty(this.typeTemplate)) {
            const isSubtype = this.typeTemplate.isSubtypeOf(type);
            if (isSubtype) {
                return isSubtype;
            }
        }
        return false;
    }
    /**
     * Can this type be coerced from
     * @param type
     */
    canCoerceFrom(type) {
        // if type no coercion needed
        if (type === this) {
            return true;
        }
        // if any type can coerce
        if (type.anyType) {
            return true;
        }
        // Are we directly a type that can be coerced
        if (this.coercibleTypes.has(type)) {
            return true;
        }
        // if we're a subtype of this type
        if (type.isSubtypeOf(this)) {
            return true;
        }
        // Are we a subtype of one of the coercible types
        for (const coercibleType of this.coercibleTypes) {
            if (type.isSubtypeOf(coercibleType)) {
                return true;
            }
        }
        return false;
    }
    /**
     * What are the type parameters of this type
     */
    getTypeParameters() {
        return [...this.typeArguments.keys()];
    }
    /**
     * Get the super type of this type
     */
    getSuperType() {
        return this.superType;
    }
    /**
     * Get the call signature of this type
     */
    getCallSignature() {
        return this.callSignature;
    }
    /**
     * Get the indexer of this type
     */
    getIndexer() {
        if (!typeGuards_1.empty(this.indexer)) {
            return this.indexer;
        }
        return this.superType && this.superType.getIndexer();
    }
    /**
     * Get the tracker of this type
     */
    getTracker() {
        return this.tracker;
    }
    /**
     * Get the assignment type of this type
     */
    getAssignmentType() {
        if (typeGuards_1.empty(this.callSignature)) {
            return this;
        }
        return this.callSignature.returns();
    }
    /**
     * Get all available coercions
     */
    getCoercions() {
        return this.coercibleTypes;
    }
    /**
     * Get all available suffixes
     */
    getSuffixes() {
        const suffixes = new Map(this.suffixes.entries());
        if (!typeGuards_1.empty(this.superType)) {
            for (const [key, value] of this.superType.getSuffixes()) {
                suffixes.set(key, value);
            }
        }
        return suffixes;
    }
    /**
     * Get the appropriate operator for this kind if it exists
     * @param kind the operator kind
     * @param other the other type if binary
     */
    getOperator(kind, other) {
        const operators = this.operators.get(kind);
        if (typeGuards_1.empty(operators)) {
            if (!typeGuards_1.empty(this.superType)) {
                return this.superType.getOperator(kind, other);
            }
            return operators;
        }
        for (const operator of operators) {
            // check if operator is unary other see if it can coerced into other operand
            if (typeGuards_1.empty(operator.secondOperand)) {
                if (typeGuards_1.empty(other)) {
                    return operator;
                }
            }
            else if (!typeGuards_1.empty(other) && operator.secondOperand.canCoerceFrom(other)) {
                return operator;
            }
        }
        return undefined;
    }
    /**
     * Get all operators of this type
     */
    getOperators() {
        return this.operators;
    }
    /**
     * Get a string representation of this type
     */
    toString() {
        const typeParameters = this.getTypeParameters();
        let typeArgumentsStr;
        if (typeParameters.length === 0) {
            typeArgumentsStr = '';
        }
        else {
            const typeArgumentStrs = [];
            for (const typeParameter of typeParameters) {
                const typeArgument = this.typeArguments.get(typeParameter);
                if (typeGuards_1.empty(typeArgument)) {
                    throw new Error(`Type argument not found for parameter ${typeParameter.toString()}` +
                        ` for type ${this.name}`);
                }
                typeArgumentStrs.push(typeArgument.toString());
            }
            typeArgumentsStr = `<${typeArgumentStrs.join(', ')}>`;
        }
        // no call signature just return type name
        if (typeGuards_1.empty(this.callSignature)) {
            return `${this.name}${typeArgumentsStr}`;
        }
        // if it's gettable or settable then show return type
        if (this.access.get || this.access.set) {
            return `${typeArgumentsStr}${this.callSignature.returns().toString()}`;
        }
        return `${typeArgumentsStr}${this.callSignature.toString()}`;
    }
    /**
     * Apply type arguments to this type
     * @param _ type arguments
     */
    apply(_) {
        return this;
    }
}
exports.Type = Type;
//# sourceMappingURL=type.js.map