"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const Stmt = __importStar(require("../parser/stmt"));
const SuffixTerm = __importStar(require("../parser/suffixTerm"));
const Expr = __importStar(require("../parser/expr"));
const path_1 = require("path");
const typeGuards_1 = require("./typeGuards");
const tokentypes_1 = require("../entities/tokentypes");
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
/**
 * Class to resolve run statements or calls to file paths
 */
class PathResolver {
    /**
     * Create an instance of that path resolve.
     * @param volume0Uri the uri associated with volume 0
     */
    constructor(volume0Uri) {
        this.volume0Uri = !typeGuards_1.empty(volume0Uri) ? vscode_uri_1.URI.parse(volume0Uri) : undefined;
    }
    /**
     * Is the resolve ready to resolve paths
     */
    ready() {
        return !typeGuards_1.empty(this.volume0Uri);
    }
    /**
     * Resolve uri to load data
     * @param caller location of caller
     * @param runPath path provided in a run statement
     */
    resolveUri(caller, runPath) {
        if (typeGuards_1.empty(runPath) || typeGuards_1.empty(this.volume0Uri)) {
            return undefined;
        }
        // get relative run path from file
        const uri = vscode_uri_1.URI.parse(caller.uri);
        // currently only support file scheme
        if (uri.scheme !== 'file') {
            return undefined;
        }
        const relativePath = path_1.relative(this.volume0Uri.toString(), path_1.dirname(caller.uri));
        // check if the scripts reads from volume 0 "disk"
        // TODO no idea what to do for ship volumes
        const [possibleVolume, ...remaining] = runPath.split('/');
        if (possibleVolume.startsWith('0:')) {
            // if of style 0:first\remaining...
            if (possibleVolume.length > 2) {
                const first = possibleVolume.slice(2);
                return this.loadData(first, ...remaining);
            }
            // else of style 0:\remaining...
            return this.loadData(...remaining);
        }
        if (relativePath === 'boot') {
            return this.loadData(possibleVolume, ...remaining);
        }
        // if no volume do a relative lookup
        return this.loadData(relativePath, possibleVolume, ...remaining);
    }
    /**
     * Creates a load data payload from a caller and path segments
     * @param caller call location
     * @param pathSegments path segments
     */
    loadData(...pathSegments) {
        if (typeGuards_1.empty(this.volume0Uri))
            return undefined;
        return vscode_uri_1.URI.file(path_1.join(this.volume0Uri.fsPath, ...pathSegments));
    }
}
exports.PathResolver = PathResolver;
/**
 * based on run type determine how to get file path
 * @param stmt run statement
 */
exports.runPath = (stmt) => {
    if (stmt instanceof Stmt.Run) {
        const { identifier } = stmt;
        switch (identifier.type) {
            case tokentypes_1.TokenType.string:
                return identifier.literal;
            case tokentypes_1.TokenType.fileIdentifier:
            case tokentypes_1.TokenType.identifier:
                return identifier.lexeme;
            default:
                return cannotLoad(stmt);
        }
    }
    // for run path variants check for literal
    const { expr } = stmt;
    if (expr instanceof Expr.Suffix) {
        if (expr.suffixTerm.atom instanceof SuffixTerm.Literal) {
            return literalPath(expr.suffixTerm.atom) || cannotLoad(stmt);
        }
    }
    return cannotLoad(stmt);
};
/**
 * determine which string to return for the filepath from literal
 * @param expr literal expression
 */
const literalPath = (expr) => {
    const { token } = expr;
    switch (token.type) {
        case tokentypes_1.TokenType.string:
            return token.literal;
        case tokentypes_1.TokenType.fileIdentifier:
            return token.lexeme;
    }
    return undefined;
};
/**
 * Normalize a filepath to the the default extension .ks if rules allow it to
 * @param uri absolute resolved path
 */
exports.normalizeExtensions = (uri) => {
    const ext = vscode_uri_1.URI.isUri(uri) ? path_1.extname(uri.fsPath) : path_1.extname(uri);
    const uriString = uri.toString();
    switch (ext) {
        case '.ks':
            return uriString;
        case '.ksm':
            return uriString.replace('.ksm', '.ks');
        case '':
            return `${uriString}.ks`;
        default:
            return undefined;
    }
};
/**
 * Provide a diagnostic when a path cannot be loaded
 * @param stmt the run statement
 */
const cannotLoad = (stmt) => vscode_languageserver_1.Diagnostic.create(stmt, 'kos-language-server cannot load runtime resolved run statements', vscode_languageserver_1.DiagnosticSeverity.Hint);
//# sourceMappingURL=pathResolver.js.map