"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const typeGuards_1 = require("./typeGuards");
const typeUtilities_1 = require("../typeChecker/typeUtilities");
const types_1 = require("../analysis/types");
const clean_1 = require("./clean");
const mapper_1 = require("./mapper");
/**
 * The default client configuration if none are available
 */
exports.defaultClientConfiguration = {
    kerbalSpaceProgramPath: undefined,
    telnetHost: '127.0.0.1',
    telnetPort: 5410,
    lspPort: 7000,
    completionCase: 'camelcase',
    trace: {
        server: {
            verbosity: 'off',
            format: 'text',
            level: 'error',
        },
    },
};
/**
 * Get the connection primitives based on the request connection type
 * @param connectionType connection type
 */
exports.getConnectionPrimitives = (program) => {
    let reader;
    let writer;
    if (program.nodeIpc) {
        reader = new vscode_languageserver_1.IPCMessageReader(process);
        writer = new vscode_languageserver_1.IPCMessageWriter(process);
    }
    else {
        reader = new vscode_languageserver_1.StreamMessageReader(process.stdin);
        writer = new vscode_languageserver_1.StreamMessageWriter(process.stdout);
    }
    writer.onError(([error, message, code]) => {
        console.log(error);
        console.log(message);
        console.log(code);
    });
    reader.onError(error => {
        console.log(error);
    });
    return {
        reader,
        writer,
    };
};
const caseMap = new Map([
    ['lowercase', 0 /* lowercase */],
    ['uppercase', 1 /* uppercase */],
    ['camelcase', 2 /* camelcase */],
    ['pascalcase', 3 /* pascalcase */],
]);
/**
 * Map a string to a case kind
 */
exports.caseMapper = mapper_1.mapper(caseMap, 'CaseKind');
const logMap = new Map([
    ['verbose', 0 /* verbose */],
    ['info', 1 /* info */],
    ['log', 2 /* log */],
    ['warn', 3 /* warn */],
    ['error', 4 /* error */],
    ['none', 5 /* none */],
]);
/**
 * Map a string to a log level
 */
exports.logMapper = mapper_1.mapper(logMap, 'LogLevel');
const symbolMap = new Map([
    [types_1.KsSymbolKind.function, vscode_languageserver_1.CompletionItemKind.Function],
    [types_1.KsSymbolKind.parameter, vscode_languageserver_1.CompletionItemKind.Variable],
    [types_1.KsSymbolKind.lock, vscode_languageserver_1.CompletionItemKind.Property],
    [types_1.KsSymbolKind.variable, vscode_languageserver_1.CompletionItemKind.Variable],
]);
/**
 * Map a ks symbol kind to a completion kind
 */
exports.symbolMapper = mapper_1.mapper(symbolMap, 'KsSymbolKind');
/**
 * Get a list of all symbols currently in scope at the given line
 * @param analyzer analyzer instance
 * @param documentPosition the current position in the document
 * @param keywordCompletions a list of keywords to always concat
 */
exports.symbolCompletionItems = (analyzer, documentPosition, keywordCompletions) => __awaiter(this, void 0, void 0, function* () {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    // get all symbols currently in scope
    const entities = yield analyzer.getScopedSymbols(position, uri);
    // generate completions
    return entities
        .map(entity => {
        const kind = exports.symbolMapper(entity.tag);
        let typeString = 'structure';
        const { tracker } = entity.name;
        if (!typeGuards_1.empty(tracker)) {
            const type = tracker.getType({ uri, range: entity.name });
            if (!typeGuards_1.empty(type)) {
                typeString = type.toString();
            }
        }
        return {
            kind,
            label: entity.name.lexeme,
            detail: `${entity.name.lexeme}: ${typeString}`,
            data: clean_1.cleanToken(entity.name),
        };
    })
        .concat(keywordCompletions)
        .map(completion => clean_1.cleanCompletion(completion));
});
/**
 * Get a list of all known suffixes give a suffix completion trigger
 * @param analyzer analyzer instance
 * @param documentPosition the current position in the document
 */
exports.suffixCompletionItems = (analyzer, documentPosition) => __awaiter(this, void 0, void 0, function* () {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    // TODO more robust method
    const token = yield analyzer.getToken({ line: position.line, character: position.character - 2 }, uri);
    if (typeGuards_1.empty(token)) {
        return [];
    }
    const type = typeUtilities_1.tokenTrackedType(token);
    // if type not found exit
    if (typeGuards_1.empty(type)) {
        return [];
    }
    // get all suffixes on the predicted type
    const suffixes = [...type.getSuffixes().values()];
    // generate completions
    return suffixes.map(suffix => ({
        kind: typeGuards_1.empty(suffix.getCallSignature())
            ? vscode_languageserver_1.CompletionItemKind.Property
            : vscode_languageserver_1.CompletionItemKind.Method,
        label: suffix.name,
        detail: `${suffix.name}: ${suffix.toString()}`,
    }));
});
/**
 * Get all symbols in the current document
 * @param analyzer analyzer instance
 * @param documentSymbol document identifier
 */
exports.documentSymbols = (analyzer, documentSymbol) => __awaiter(this, void 0, void 0, function* () {
    const { uri } = documentSymbol.textDocument;
    const entities = yield analyzer.getAllFileSymbols(uri);
    return entities.map(entity => {
        let kind = undefined;
        switch (entity.tag) {
            case types_1.KsSymbolKind.function:
                kind = vscode_languageserver_1.SymbolKind.Function;
                break;
            case types_1.KsSymbolKind.parameter:
                kind = vscode_languageserver_1.SymbolKind.Variable;
                break;
            case types_1.KsSymbolKind.lock:
                kind = vscode_languageserver_1.SymbolKind.Object;
                break;
            case types_1.KsSymbolKind.variable:
                kind = vscode_languageserver_1.SymbolKind.Variable;
                break;
            default:
                throw new Error('Unknown symbol type');
        }
        return {
            kind,
            name: entity.name.lexeme,
            location: clean_1.cleanLocation({
                uri: entity.name.uri || uri,
                range: entity.name,
            }),
            containerName: 'example',
        };
    });
});
/**
 * The default signature if non can be provided
 */
exports.defaultSignature = () => ({
    signatures: [],
    activeParameter: null,
    activeSignature: null,
});
/**
 * Convert parser error to diagnostic
 * @param error parser error
 * @param uri uri string
 */
exports.parseToDiagnostics = (error, uri) => {
    return {
        uri,
        severity: vscode_languageserver_1.DiagnosticSeverity.Error,
        range: { start: error.start, end: error.end },
        message: error.message,
        source: 'kos-language-server',
    };
};
/**
 * convert resolver error to diagnostic
 * @param error diagnostic
 * @param uri uri string
 */
exports.addDiagnosticsUri = (error, uri) => {
    return Object.assign({ uri }, error);
};
//# sourceMappingURL=serverUtils.js.map